"""Diagonal assembly."""

from bempp_cl.api.assembly import assembler as _assembler


class DiagonalAssembler(_assembler.AssemblerBase):
    """Implementation of a diagonal assembler for integral operators."""

    # pylint: disable=useless-super-delegation
    def __init__(self, domain, dual_to_range, parameters=None):
        """Create a diagonal assembler instance."""
        super().__init__(domain, dual_to_range, parameters)

    def assemble(self, operator_descriptor, device_interface, precision, *args, **kwargs):
        """Diagonal assembly of the integral operator."""
        from bempp_cl.api.assembly.discrete_boundary_operator import DiagonalOperator

        if self.domain.requires_dof_transformation or self.dual_to_range.requires_dof_transformation:
            raise ValueError("Spaces that require dof transformations not supported for diagonal assembly.")

        if self.domain != self.dual_to_range:
            raise ValueError("Only identical spaces currently supported for diagonal assembly.")

        if (
            self.domain.identifier == self.dual_to_range.identifier and self.dual_to_range.identifier == "p1_continuous"
        ) or (
            self.domain.identifier == self.dual_to_range.identifier
            and self.dual_to_range.identifier == "p0_discontinuous"
        ):
            values = operator_descriptor.singular_part.weak_form().to_sparse().diagonal()
        else:
            raise ValueError(
                "Only spaces of type 'p0_discontinuous' or 'p1_continuous' supported for diagonal assembly."
            )

        return DiagonalOperator(
            values,
            shape=(self.dual_to_range.global_dof_count, self.domain.global_dof_count),
        )



"""Bempp direct solver interface."""

# pylint: disable=invalid-name


def compute_lu_factors(A):
    """
    Precompute the LU factors of a dense operator A.

    This function returns a tuple of LU factors of A.
    This tuple can be used in the `lu_factor` attribute
    of the lu function so that the LU decomposition is
    not recomputed at each call to lu.

    """
    from bempp_cl.api import as_matrix
    from scipy.linalg import lu_factor

    return lu_factor(as_matrix(A.weak_form()))


def lu(A, b, lu_factor=None):
    """Perform an LU solve.

    This function takes an operator and a grid function,
    converts the operator into a dense matrix and solves
    the system via LU decomposition. The result is again
    returned as a grid function.

    Parameters
    ----------
    A : bempp_cl.api.BoundaryOperator
         The left-hand side boundary operator
    b : bempp_cl.api.GridFunction
         The right-hand side grid function
    lu_decomp : tuple
         Optionally pass the tuple (lu, piv)
         obtained by the scipy method scipy.linalg.lu_factor

    """
    from bempp_cl.api import GridFunction
    from scipy.linalg import solve, lu_solve
    from bempp_cl.api.assembly.blocked_operator import BlockedOperatorBase
    from bempp_cl.api.assembly.blocked_operator import projections_from_grid_functions_list
    from bempp_cl.api.assembly.blocked_operator import grid_function_list_from_coefficients

    if isinstance(A, BlockedOperatorBase):
        vec = projections_from_grid_functions_list(b, A.dual_to_range_spaces)
        if lu_factor is not None:
            sol = lu_solve(lu_factor, vec)
        else:
            mat = A.weak_form().to_dense()
            sol = solve(mat, vec)
        return grid_function_list_from_coefficients(sol, A.domain_spaces)
    else:
        vec = b.projections(A.dual_to_range)
        if lu_factor is not None:
            sol = lu_solve(lu_factor, vec)
        else:
            mat = A.weak_form().to_dense()
            sol = solve(mat, vec)
        return GridFunction(A.domain, coefficients=sol)



"""
Functions to import / export datasets (mostly import for now)
"""

import numpy as np
from copy import copy
from os import listdir
from os.path import isfile, join
import re
import pyuff
from scipy.signal.windows import tukey


def load_impulse(folderPath, isConvertedFromMAT=False):
    """
    Import impulse response data from CSV files.

    Parameters
    ----------
    folderPath : str
        Path to the folder where IR.csv files are stored.

    isConvertedFromMAT : bool, optional
        If True, indicates that the data has been converted from .mat to .csv format,
        which may affect the number of rows to skip when loading data. Default is False.

    Returns
    -------
    theta : numpy array
        Array of angles in degrees corresponding to the imported impulse responses.

    IR_MATRIX : numpy array
        Impulse response matrix of shape (len(theta), len(IR)).

    Notes
    -----
    This function imports impulse response data from CSV files stored in the specified
    'folderPath'. The function returns a tuple containing two numpy arrays: 'theta' and
    'IR_MATRIX'. The 'theta' array holds the angles in degrees corresponding to the
    imported impulse responses. The 'IR_MATRIX' array holds the actual impulse response
    data in a matrix format, where each row represents an angle and each column represents
    a time step in the impulse response.

    If the 'isConvertedFromMAT' parameter is set to True, the function adjusts the number of
    rows to skip when loading data to accommodate the conversion from .mat to .csv format.
    """
    files = [f for f in listdir(folderPath) if isfile(join(folderPath, f))]
    files = [f for f in files if "IR" in f]

    # load file to get length of time data
    if isConvertedFromMAT is False:
        dummyData = np.loadtxt(join(folderPath, files[0]), skiprows=4, delimiter=',')
        lenData = len(dummyData[:, 0])
        skiprow = 4
    elif isConvertedFromMAT is True:
        dummyData = np.loadtxt(join(folderPath, files[0]), skiprows=1, delimiter=',')
        lenData = len(dummyData[:, 0])
        skiprow = 1

    # build angle array
    pattern = r'[\d\.]+'
    numerical_values = re.findall(pattern, files[0][:-4])

    where_theta = files[0].index(numerical_values[0])

    theta = np.zeros(len(files))
    for t in range(len(theta)):
        # print("where is theta: ", where_theta)
        # print("theta: ", theta[t])
        theta[t] = float(files[t][where_theta:where_theta + 3])
    theta = np.sort(theta)

    # build IR matrix
    IR_MATRIX = np.zeros([len(theta), lenData])
    for t in range(len(theta)):
        filepath = join(folderPath, files[0][:where_theta] + str(theta[t]) + '_' + 'IR.csv')
        data = np.loadtxt(filepath, skiprows=skiprow, delimiter=',')
        time = data[:, 0]
        signal = data[:, 1]
        IR_MATRIX[t, :] = signal
    return theta, time, IR_MATRIX


def processIR(IR_matrix, Fs, minTime, maxTime, alpha=0.02):
    indMin = int(minTime * Fs)
    indMax = int(maxTime * Fs)
    indTot = indMax - indMin
    window = tukey(int(indTot), alpha)

    IR_out = np.zeros([len(IR_matrix), indTot])

    for i in range(len(IR_matrix)):
        IR_out[i, :] = IR_matrix[i, indMin:indMax] * window
    return IR_out

def processFRF(IR_matrix, Fs, Nfft=None):
    if Nfft == None:
        Nfft = np.shape(IR_matrix)[1]
    FRF_out = np.zeros([len(IR_matrix), Nfft], dtype=complex)
    for i in range(len(IR_matrix)):
        FRF_out[i, :] = np.fft.fft(IR_matrix[i, :], Nfft)
    freq = np.arange(Nfft) * Fs / Nfft
    return freq, FRF_out

# load acceleration data from uff file
def loadUFF(uffFile):
    """
    Load acceleration / velocity data from a UFF file and return frequency axis, acceleration data,
    and point positions in space.

    Parameters
    ----------
    uffFile : str
        Path to the *.uff file.

    Returns
    -------
    uff_data : tuple
        A tuple containing the following elements:
        - freq : numpy array
            Frequency axis corresponding to the loaded acceleration data.
        - a_matrix : numpy array, shape (n_acc_points, nfft)
            Matrix of complex-valued acceleration data.
        - X : numpy array, shape (nPoints, 3)
            Matrix containing point positions in 3D space (x, y, z).

    Notes
    -----
    This function loads acceleration data from a UFF (Universal File Format) file and extracts
    relevant information. The function returns a tuple 'uff_data' containing three elements:

    - 'freq': A numpy array representing the frequency axis corresponding to the loaded
      acceleration data.
    - 'a_matrix': A numpy array with shape (n_acc_points, nfft), containing complex-valued
      acceleration data. Each row corresponds to a different point in space.
    - 'X': A numpy array with shape (nPoints, 3) representing the point positions in 3D space.
      Each row represents a point, and the columns correspond to the coordinates (x, y, z).
    """
    file = pyuff.UFF(uffFile)
    data = file.read_sets()

    # create acceleration matrix
    nData = len(data)

    point_count_acc = 0
    point_count_vel = 0
    index_acc = []
    index_vel = []
    index_58 = []
    for i in range(nData):
        if data[i]['type'] == 58:
            if data[i]['ordinate_axis_lab'] == 'Acceleration' and data[i]['orddenom_axis_lab'] == 'Voltage':
                # print('here')
                point_count_acc += 1
                index_acc.append(i)
            if data[i]['ordinate_axis_lab'] == 'Velocity':
                point_count_vel += 1
                index_vel.append(i)
            index_58.append(i)

    nfft = len(data[index_58[0]]['data'])
    a_matrix = np.zeros([point_count_acc, nfft], dtype=complex)
    v_matrix = np.zeros([point_count_vel, nfft], dtype=complex)
    # node_id = np.zeros(len(index_58))
    for i, val in enumerate(index_acc):
        a_matrix[i, :] = data[val]['data']
    for i, val in enumerate(index_vel):
        v_matrix[i, :] = data[val]['data']
        # node_id[i] = data[val]['rsp_node']

    freq = data[index_58[0]]['x']

    # space data
    point_count = 0
    index_2411 = []
    for i in range(len(data)):
        if data[i]['type'] == 2411:
            point_count += 1
            index_2411.append(i)

    x, y, z = data[index_2411[0]]['x'], \
        data[index_2411[0]]['y'], \
        data[index_2411[0]]['z']
    X = np.zeros([len(x), 3])
    X[:, 0] = x
    X[:, 1] = y
    X[:, 2] = z

    # output data
    uff_data = (freq, a_matrix, v_matrix, X)
    return uff_data


def loadREW_FRF(file, skiprows=14):
    data = np.loadtxt(file, skiprows=skiprows)
    F = data[:, 0]
    Hf = data[:, 1]
    Ht = data[:, 2]
    return F, Hf, Ht


"""Define the default parameters object."""

# pylint: disable=too-few-public-methods


class _Verbosity(object):
    """Verbosity Parameters."""

    def __init__(self):
        """Initialize verbosity parameters."""
        self.extended_verbosity = False


class _Output(object):
    """Output parameters."""

    def __init__(self):
        """Initialize output parameters."""
        self.gmsh_use_binary = True


class _Quadrature(object):
    """Quadrature orders."""

    def __init__(self):
        """Iniitalize quadrature parameters."""
        self.regular = 4
        self.singular = 4


class _Fmm(object):
    """Fmm options."""

    def __init__(self):
        """Iniitalize FMM parameters."""
        self.expansion_order = 5
        self.depth = 4
        self.ncrit = 400
        self.near_field_representation = "evaluate"
        self.debug = False
        self.dense_evaluation = False


class _DenseAssembly(object):
    """Dense assembly options."""

    def __init__(self):
        """Iniitalize dense assembly parameters."""
        self.workgroup_size_multiple = 2


class _Assembly(object):
    """Assembly options."""

    def __init__(self):
        """Iniitalize assembly parameters."""
        self.dense = _DenseAssembly()
        self.always_promote_to_double = False
        self.discretization_type = "galerkin"


class DefaultParameters(object):
    """Default parameters for Bempp."""

    def __init__(self):
        """Initialize parameters."""
        self.verbosity = _Verbosity()
        self.output = _Output()
        self.quadrature = _Quadrature()
        self.assembly = _Assembly()
        self.fmm = _Fmm()



"""Bempp: Boundary element method Python package.

Bempp-cl is an open-source boundary element method library that
can be used to assemble all the standard integral kernels for
Laplace, Helmholtz, modified Helmholtz, and Maxwell problems.

Documentation can be found at https://bempp.com/documentation/index.html
"""

from .version import __version__



"""
Helper functions for BEM and point-source methods

"""
import numpy as np
import bempp_cl.api


#%% Green-function and monopole 
def incidenceCoeff(xSource, Grid, domain, dofs="centroids"):
    """
    Get cos(angle) of source incidence to grid normals. In combination to -1jk, will 
    give the normal derivative (not exactly sure if this is the right term).

    Parameters
    ----------
    xSource : ndarray
        Array of source position (Nsource, 3).
    Grid : bempp-cl grid object
        System grid.
    domain : str
        Domain of simulation. Either "interior" or "exterior".
    dofs : str, optional
        Which DOFs to select. The default is "centroids".

    Returns
    -------
    n0 : TYPE
        DESCRIPTION.

    """
    if domain == "interior":
        domain_operator = -1
    else:
        domain_operator = 1
    Nvert = Grid.vertices.shape[1]
    Norm  = domain_operator * Grid.normals.T
    Vert  = Grid.vertices
    Cent  = Grid.centroids.T
    Ncent = Grid.centroids.shape[0]
    Ns    = len(xSource)

    if dofs == "centroids":
        DOF = Cent
        n0 = np.zeros(Ncent)
        dir_vect = np.zeros([3, Ncent, Ns])
        nDOF = Ncent
    
    elif dofs == "vertices":
        DOF = Vert
        n0 = np.zeros(Nvert)
        dir_vect = np.zeros([3, Nvert, Ns])
        nDOF = Nvert
    
    for s in range(Ns):
        dir_vect[0, :, s] = DOF[0, :] - xSource[s, 0]
        dir_vect[1, :, s] = DOF[1, :] - xSource[s, 1]
        dir_vect[2, :, s] = DOF[2, :] - xSource[s, 2]
    
    for i in range(3):
        for s in range(Ns):
            norm_tmp = (dir_vect[0, i, s]**2 + 
                        dir_vect[1, i, s]**2 + dir_vect[2, i, s]**2)**0.5
            dir_vect[:, i, s] /= norm_tmp #np.max(np.abs(dir_vect[:, i, s]))  
    
    
    for dof in range(nDOF):
        for s in range(len(xSource)): 
            n0[dof] += (np.dot(dir_vect[:, dof, s], Norm[:, dof]) / 
                    np.linalg.norm(dir_vect[:, dof, s]) / 
                    np.linalg.norm(Norm[:, dof]))
    return n0

def greenMonopole(source, grid, k, position="centroids"):
    """
    Compute Green function between a (NSource, 3) source array and bempp grid object 
    (either centroids or vertices - DOFs)

    Parameters
    ----------
    source : ndarray,
        Position of monopole sources in space.
    grid : bempp-cl grid object,
        Grid containting mesh information.
    k : float
        Wavenumber.
    position : str, optional
        Which DOFs. Can be "centroids" or "vertices". The default is "centroids".

    Returns
    -------
    G : ndarray
        Vector of Green functions.

    """
    if position=="centroids":
        X_grid = grid.centroids
    else:
        X_grid = grid.vertices
    G = np.zeros(len(X_grid), dtype=complex)
    for i in range(len(source)):
        dist = ((source[i, 0]-X_grid[:, 0])**2 + 
                (source[i, 1]-X_grid[:, 1])**2 +
                (source[i, 2]-X_grid[:, 2])**2)**0.5
        G += np.exp(1j*k*dist) / (4*np.pi*dist)
    return G

def buildGridFunction(space, *args):
    dof = space.grid_dof_count
    coeff = np.ones(dof, dtype=complex)
    for arg in args:
        coeff *= arg
    grid_fun = bempp_cl.api.GridFunction(space, coefficients=coeff)
    return grid_fun

def element2vertice_pressure(grid, p_mesh):
    """
    "Transpose" element pressure to vertex pressure.

    Parameters
    ----------
    grid : bempp-cl grid object
        Simulation grid.
    p_mesh : ndarray of gridfunctions
        Result of the pressure over the system's mesh.

    Returns
    -------
    vertice_pressure : ndarray of gridfunctions
        Transposed element to vertex pressure.
    """

    vertices = grid.vertices
    elements = grid.elements
    N = vertices.shape[1]
    M = elements.shape[1]
    
    # Step 1: Initialize an array to accumulate pressure per vertex
    vertice_pressure = np.empty(p_mesh.shape, dtype="object")
    space_vert = bempp_cl.api.function_space(grid, "P", 1)
    
    # Step 2: Loop through each element and distribute pressure
    for f in range(p_mesh.shape[0]):
        for rs in range(p_mesh.shape[1]):
            coeff = np.zeros((len(p_mesh), N), dtype=complex)
            vertex_count = np.zeros(N)  # To count contributions per vertex
            for i in range(M):
                for j in range(3):  # Each element has 3 vertices
                    v_idx = elements[j, i]  # Get vertex index
                    coeff[f, v_idx] += p_mesh[f, rs].coefficients[i]  # sum elements to a single vertex position
                    vertex_count[v_idx] += 1  # Count contributions
        
            nonzero_mask = vertex_count > 0
            coeff[f, nonzero_mask] /= vertex_count[nonzero_mask] 
            vertice_pressure[f, rs] = bempp_cl.api.GridFunction(space_vert, 
                                                             coefficients=coeff[f, :])
    
    # for f in range(p_mesh.shape[0]):
    #     for rs in range(p_mesh.shape[1]):
    #         projection_coeff = p_mesh[f, rs].projections(space_vert)
    #         vertice_pressure[f, rs] = bempp_cl.api.GridFunction(space_vert, coefficients=projection_coeff)
    return vertice_pressure

 
#%% Mesh-related
def mirror_mesh(grid_init, boundary_conditions):
    bc = boundary_conditions
    size_factor = 1
    grid_tot = grid_init
    if bc is not None:
        for item in bc:
            boundary = bc[item]
            if boundary["type"] == "infinite_baffle":
                offset = boundary["offset"]
                vertices = np.copy(grid_tot.vertices)
                elements = np.copy(grid_tot.elements)
                if item in ["x", "X"]:
                    if offset != 0:
                        vertices[0, :] = 2*offset - vertices[0, :]
                        elements[[2, 0], :] = elements[[0, 2], :]
                    else:
                        vertices[0, :] = vertices[0, :]
                        elements[[2, 0], :] = elements[[0, 2], :]                    
                elif item in ["y", "Y"]:
                    if offset != 0 :
                        vertices[1, :] = 2*offset - vertices[1, :]
                        elements[[2, 0], :] = elements[[0, 2], :]
                    else:
                        vertices[1, :] = vertices[1, :]
                        elements[[2, 0], :] = elements[[0, 2], :]    
                elif item in ["z", "Z"]:
                    if offset is not False:
                        vertices[2, :] = 2*offset - vertices[2, :]
                        elements[[2, 0], :] = elements[[0, 2], :]
                    else:
                        vertices[2, :] = vertices[2, :]
                        elements[[2, 0], :] = elements[[0, 2], :]               
                else:
                    print("{} not infinite baffle.".format(item))
                grid_mirror = bempp_cl.api.Grid(vertices, elements, domain_indices=grid_tot.domain_indices)
                grid_tot = bempp_cl.api.grid.union([grid_tot, grid_mirror],
                                                [grid_tot.domain_indices, grid_mirror.domain_indices])
                size_factor *= 2
    return grid_tot, size_factor

#%% Admittance
def getSurfaceAdmittance(absorbingSurface, surfaceImpedance, freq, spaceP, c_0, rho_0):
    """
    Compute the total single layer coefficients linked to surfaces impedance.
    :param absorbingSurface:
    :param surfaceImpedance:
    :param freq:
    :param spaceP:
    :return:
    """
    Nfft       = len(freq)
    absSurf_in = np.array(absorbingSurface)
    surfImp_in = surfaceImpedance
    Nsurf      = len(absSurf_in)             # number of absorbing surfaces
    grid       = spaceP.grid
    dofCount   = spaceP.grid_dof_count

    if absSurf_in.shape[0] == 0 :
        admittanceMatrix = None
    else:
        admittanceMatrix = np.zeros([dofCount, Nfft], dtype=complex) #np.ones([dofCount, Nfft], dtype=complex) * 2.5e-4 # corresponds to 0.05% damping
        for surf in range(Nsurf):
            tmp_surf = absSurf_in[surf]  # current surface on which we apply admittance coefficients
            vertex, _ = get_group_points(grid, tmp_surf)
            for f in range(Nfft):
                try:
                    Yn = rho_0 * c_0 / surfImp_in[surf][f] # rho_0 * c_0
                except:
                    Yn = rho_0 * c_0 / surfImp_in[surf] # rho_0 * c_0
                admittanceMatrix[vertex, f] = np.ones(len(vertex)) * Yn
    return admittanceMatrix 



def get_group_points(grid, group_number):
    domain_indices = grid.domain_indices
    elements       = grid.elements
    
    # Find the indices where domain_indices match the group_number
    group_indices = np.where(domain_indices == group_number)  # indices of support elements

    # Get the corresponding columns of elements
    group_points = elements[:, group_indices]  # segments (and NOT points!) part of group_number

    # Reshape to a 1D array and use unique to get unique values
    unique_points   = np.unique(group_points)

    return unique_points, group_indices






"""Implementation of potential operators."""


class DensePotentialAssembler(object):
    """Implementation of a potential assembler."""

    # pylint: disable=useless-super-delegation
    def __init__(
        self,
        space,
        operator_descriptor,
        points,
        device_interface,
        parameters=None,
    ):
        """Create a dense assembler instance."""
        from bempp_cl.core.dispatcher import potential_dispatcher

        implementation = potential_dispatcher(
            device_interface,
            space.localised_space,
            operator_descriptor,
            points,
            parameters,
        )

        self.space = space
        kernel_dimension = operator_descriptor.kernel_dimension

        def potential_evaluator(x):
            """Evaluate the potential."""
            x_transformed = self.space.map_to_full_grid @ (self.space.dof_transformation @ x)
            result = implementation(x_transformed)
            return result.reshape([kernel_dimension, -1], order="F")

        self._evaluator = potential_evaluator

    def evaluate(self, x):
        """Call the potential evaluator."""
        return self._evaluator(x)



"""Interfaces to Maxwell operators."""

import numpy as _np

from bempp_cl.api.operators.boundary import common as _common


def electric_field(
    domain,
    range_,
    dual_to_range,
    wavenumber,
    parameters=None,
    assembler="default_nonlocal",
    device_interface=None,
    precision=None,
):
    """Assemble the electric field boundary operator."""
    if domain.identifier != "rwg0":
        raise ValueError("Domain space must be an RWG type function space.")

    if dual_to_range.identifier != "snc0":
        raise ValueError("Dual to range space must be an SNC type function space.")

    return _common.create_operator(
        "maxwell_electric_field_boundary",
        domain,
        range_,
        dual_to_range,
        parameters,
        assembler,
        [_np.real(wavenumber), _np.imag(wavenumber)],
        "helmholtz_single_layer",
        "maxwell_electric_field",
        device_interface,
        precision,
        True,
    )


def magnetic_field(
    domain,
    range_,
    dual_to_range,
    wavenumber,
    parameters=None,
    assembler="default_nonlocal",
    device_interface=None,
    precision=None,
):
    """Assemble the magnetic field boundary operator."""
    if domain.identifier != "rwg0":
        raise ValueError("Domain space must be an RWG type function space.")

    if dual_to_range.identifier != "snc0":
        raise ValueError("Dual to range space must be an SNC type function space.")

    return _common.create_operator(
        "maxwell_magnetic_field_boundary",
        domain,
        range_,
        dual_to_range,
        parameters,
        assembler,
        [_np.real(wavenumber), _np.imag(wavenumber)],
        "helmholtz_single_layer",
        "maxwell_magnetic_field",
        device_interface,
        precision,
        True,
    )


def multitrace_operator(
    grid,
    wavenumber,
    epsilon_r=1,
    mu_r=1,
    target=None,
    space_type="magnetic_dual",
    parameters=None,
    assembler="default_nonlocal",
    device_interface=None,
    precision=None,
):
    """
    Simplified version of multitrace operator assembly.

    Parameters
    ----------
    grid : Grid
        Bempp grid object.
    wavenumber : complex
        A real or complex wavenumber
    epsilon_r : float
        Relative permittivity with respect to vacuum.
    mu_r : float
        Relative permeability with respect to vacuum.
    target : Grid
        The grid for the range spaces. If target is None then
        target is set to the input grid (that is the domain
        grid).
    space_type : string
        One of "all_rwg", "all_bc", "magnetic_dual" (default),
        "electric_dual". These lead to the following
        choices of space, range, and dual_to_range:
        default - (RWG, RWG), (BC, BC), (SNC, SNC)
        all_dual - (BC, BC), (RWG, RWG), (RBC, RBC)
        magnetic_dual - (RWG, BC), (RWG, BC), (RBC, SNC)
        electric_dual - (BC, RWG), (BC, RWG), (SNC, RBC)
    parameters : Parameters
        An optional parameters object.
    assembler : string
        The assembler type.
    device_interface : DeviceInterface
        The device interface object to be used.
    precision : string
        Either "single" or "double" for single or
        double precision mode.

    Output
    ------
    The Maxwell multitrace operator of the form
    [[M, E], [-E, M]], where M represens the magnetic
    and E the electric field boundary operators in
    the respective spaces defined through space_type.
    Note that the operators in the first and second
    row have different discretisations depending on
    the type of spaces used.

    """
    import bempp_cl.api

    if space_type == "all_rwg":
        rwg = bempp_cl.api.function_space(grid, "RWG", 0)
        bc = bempp_cl.api.function_space(grid, "BC", 0)
        snc = bempp_cl.api.function_space(grid, "SNC", 0)

        if target is not None:
            bc_target = bempp_cl.api.function_space(target, "BC", 0)
            snc_target = bempp_cl.api.function_space(target, "SNC", 0)
        else:
            bc_target = bc
            snc_target = snc

        domain = [rwg, rwg]
        range_ = [bc_target, bc_target]
        dual_to_range = [snc_target, snc_target]
    elif space_type == "all_bc":
        bc = bempp_cl.api.function_space(grid, "BC", 0)
        rwg = bempp_cl.api.function_space(grid, "RWG", 0)
        rbc = bempp_cl.api.function_space(grid, "RBC", 0)

        if target is not None:
            rwg_target = bempp_cl.api.function_space(target, "RWG", 0)
            rbc_target = bempp_cl.api.function_space(target, "RBC", 0)
        else:
            rwg_target = rwg
            rbc_target = rbc

        domain = [bc, bc]
        range_ = [rwg_target, rwg_target]
        dual_to_range = [rbc_target, rbc_target]

    elif space_type == "electric_dual":
        rwg = bempp_cl.api.function_space(grid, "RWG", 0)
        snc = bempp_cl.api.function_space(grid, "SNC", 0)
        bc = bempp_cl.api.function_space(grid, "BC", 0)
        rbc = bempp_cl.api.function_space(grid, "RBC", 0)

        if target is not None:
            rwg_target = bempp_cl.api.function_space(target, "RWG", 0)
            snc_target = bempp_cl.api.function_space(target, "SNC", 0)
            bc_target = bempp_cl.api.function_space(target, "BC", 0)
            rbc_target = bempp_cl.api.function_space(target, "RBC", 0)
        else:
            rwg_target = rwg
            snc_target = snc
            bc_target = bc
            rbc_target = rbc

        domain = [rwg, bc]
        range_ = [rwg_target, bc_target]
        dual_to_range = [rbc_target, snc_target]

    elif space_type == "magnetic_dual":
        rwg = bempp_cl.api.function_space(grid, "RWG", 0)
        snc = bempp_cl.api.function_space(grid, "SNC", 0)
        bc = bempp_cl.api.function_space(grid, "BC", 0)
        rbc = bempp_cl.api.function_space(grid, "RBC", 0)

        if target is not None:
            rwg_target = bempp_cl.api.function_space(target, "RWG", 0)
            snc_target = bempp_cl.api.function_space(target, "SNC", 0)
            bc_target = bempp_cl.api.function_space(target, "BC", 0)
            rbc_target = bempp_cl.api.function_space(target, "RBC", 0)
        else:
            rwg_target = rwg
            snc_target = snc
            bc_target = bc
            rbc_target = rbc

        domain = [bc, rwg]
        range_ = [bc_target, rwg_target]
        dual_to_range = [snc_target, rbc_target]

    else:
        raise ValueError("space_type must be one of 'all_rwg', 'all_dual', 'electric_dual', 'magnetic_dual'")
    return _multitrace_operator_impl(
        domain,
        range_,
        dual_to_range,
        wavenumber,
        epsilon_r=epsilon_r,
        mu_r=mu_r,
        parameters=parameters,
        assembler=assembler,
        device_interface=device_interface,
        precision=precision,
    )


def _multitrace_operator_impl(
    domain,
    range_,
    dual_to_range,
    wavenumber,
    epsilon_r=1,
    mu_r=1,
    parameters=None,
    assembler="default_nonlocal",
    device_interface=None,
    precision=None,
):
    from bempp_cl.api.assembly.blocked_operator import BlockedOperator

    rho = _np.sqrt(epsilon_r / mu_r)

    magnetic1 = magnetic_field(
        domain[0],
        range_[0],
        dual_to_range[0],
        wavenumber,
        parameters=parameters,
        assembler=assembler,
        device_interface=device_interface,
        precision=precision,
    )

    magnetic2 = magnetic_field(
        domain[1],
        range_[1],
        dual_to_range[1],
        wavenumber,
        parameters=parameters,
        assembler=assembler,
        device_interface=device_interface,
        precision=precision,
    )

    electric1 = electric_field(
        domain[1],
        range_[0],
        dual_to_range[0],
        wavenumber,
        parameters=parameters,
        assembler=assembler,
        device_interface=device_interface,
        precision=precision,
    )

    electric2 = electric_field(
        domain[0],
        range_[1],
        dual_to_range[1],
        wavenumber,
        parameters=parameters,
        assembler=assembler,
        device_interface=device_interface,
        precision=precision,
    )

    blocked = BlockedOperator(2, 2)
    blocked[0, 0] = magnetic1
    blocked[0, 1] = (1.0 / rho) * electric1
    blocked[1, 0] = -rho * electric2
    blocked[1, 1] = magnetic2

    return blocked


class _OsrcMtE:
    """Implementation of the OSRC DtN operator."""

    def __init__(
        self, type, domains, ranges, dual_to_ranges, parameters, operator_options, device_interface=None, precision=None
    ):
        from bempp_cl.api.operators import OperatorDescriptor

        self._device_interface = device_interface
        self.type = type
        self._operator_descriptor = OperatorDescriptor(
            "osrc_mte",
            operator_options,
            "vectorial_laplace_beltrami",
            "default_sparse",
            precision,
            True,
            None,
            1,
        )

        self.domains_ = domains
        self.ranges_ = ranges
        self.dual_to_ranges_ = dual_to_ranges
        self.coeff = 0

    @property
    def descriptor(self):
        """Operator descriptor."""
        return self._operator_descriptor

    def _matvec1(self, v):
        import itertools

        res = 0
        rhs = list(itertools.chain(*[v.tolist(), _np.zeros(self.pi[0].shape[0] - v.shape[0]).tolist()]))
        for element in self.pi:
            res -= element.matvec(rhs)[0 : self.domains_[0].global_dof_count]
        res = self.mass * res + self.pade_coeffs[3] * v
        return self.lambda_2_inv * res

    def _matvec2(self, v):
        res = self.coeff * self.mass * v + self.pade_coeffs[3] * v
        return self.lambda_2_inv * res

    def _assemble(self):
        """Assemble the operator."""
        from bempp_cl.api.operators.boundary.sparse import lambda_1
        from bempp_cl.api.operators.boundary.sparse import lambda_2
        from bempp_cl.api.operators.boundary.sparse import mte_operators
        from bempp_cl.api.assembly.discrete_boundary_operator import InverseSparseDiscreteBoundaryOperator
        from scipy.sparse.linalg import LinearOperator

        wavenumber, npade, theta, damped_wavenumber = self.descriptor.options

        if damped_wavenumber is None:
            dk = wavenumber + 1.0j * 0.39 * wavenumber ** (1.0 / 3) * _np.sqrt(2) ** (2.0 / 3)
        else:
            dk = damped_wavenumber

        mte_op = mte_operators(self.domains_, self.ranges_, self.dual_to_ranges_, dk)

        self.lambda_2_inv = InverseSparseDiscreteBoundaryOperator(lambda_2(mte_op).weak_form())

        self.pade_coeffs = _common.pade_coeffs(npade, theta)
        self.mass = mte_op[1].weak_form()
        if self.type == 1:
            self.pi = []
            for i in range(npade):
                self.pi.append(
                    (self.pade_coeffs[1][i] / self.pade_coeffs[2][i]) * lambda_1(mte_op, self.pade_coeffs[2][i], dk)
                )
            return LinearOperator(self.lambda_2_inv.shape, matvec=self._matvec1)
        else:
            for j in range(int(_np.floor(4.0 * npade / 5)), npade):
                self.coeff += self.pade_coeffs[1][j]
            return LinearOperator(self.lambda_2_inv.shape, matvec=self._matvec2)


def osrc_mte(
    domains,
    ranges,
    dual_to_ranges,
    wavenumber,
    npade=2,
    theta=_np.pi / 2.0,
    type=1,
    damped_wavenumber=None,
    parameters=None,
    device_interface=None,
    precision=None,
):
    """Assemble the OSRC approximation to the NtD operator."""
    if domains[0].identifier != "snc0":
        raise ValueError("Domain space must be a SNC type function space.")

    if domains[1].shapeset.identifier != "p1_discontinuous":
        raise ValueError("Domain space must be of type 'p1_discontinuous'.")

    if ranges[0].identifier != "snc0":
        raise ValueError("Domain space must be a SNC type function space.")

    if ranges[1].shapeset.identifier != "p1_discontinuous":
        raise ValueError("Domain space must be of type 'p1_discontinuous'.")

    if dual_to_ranges[0].identifier != "snc0":
        raise ValueError("Domain space must be a SNC type function space.")

    if dual_to_ranges[1].shapeset.identifier != "p1_discontinuous":
        raise ValueError("Domain space must be of type 'p1_discontinuous'.")

    return _OsrcMtE(
        type,
        domains,
        ranges,
        dual_to_ranges,
        parameters,
        [wavenumber, npade, theta, damped_wavenumber],
        device_interface,
        precision,
    )



#!/usr/bin/python3
import tkinter as tk
import tkinter.ttk as ttk
from tkinter import filedialog
from electroacPy.speakerSim.filterDesign import xovers
from electroacPy.loudspeakerSystem import updateResults, apply_Velocity_From_EAD, apply_Velocity_From_PLV, apply_Velocity_From_SPKBOX
from electroacPy.io import load, save
import numpy as np
from matplotlib.figure import Figure
from matplotlib.backends.backend_tkagg import (FigureCanvasTkAgg,
                                               NavigationToolbar2Tk)
import generalToolbox as gtb


class XoverAppApp:
    def __init__(self, master=None):
        # General properties
        self.data_path = ""
        self.pMic = {}
        self.frequency = 0
        self.observation = {}
        self.angle = {}
        self.xovers = {}
        self.canvas_widget = {}
        self.fig = {}
        self.canvas = {}
        self.fig_sum = {}
        self.canvas_sum = {}
        self.dat = None
        self.study = None

        # build ui
        self.toplevel1 = tk.Tk() if master is None else tk.Toplevel(master)
        self.toplevel1.configure(height=200, width=200)
        self.toplevel1.title("ElectroacPy - crossover designer")
        self.frame1 = ttk.Frame(self.toplevel1, name="frame1")
        self.frame1.configure(height=720, width=1250)
        self.panedWindow = ttk.Panedwindow(self.frame1, orient='horizontal')
        self.panedWindow.configure(height=720, width=1250)
        self.notebook = ttk.Notebook(self.panedWindow, name="notebook")
        self.notebook.configure(height=720, width=250)

        # =======================
        # DATASET TAB
        self.frame_import = ttk.Frame(self.notebook, name="frame_import")
        self.frame_import.configure(height=720, width=250)
        self.button_load_data = ttk.Button(
            self.frame_import, name="button_load_data")
        self.button_load_data.configure(text='Import Data')
        self.button_load_data.place(
            anchor="nw", relx=0.05, rely=0.05, x=0, y=0)
        self.button_load_data.configure(command=self.load_data)

        self.button_export = ttk.Button(
            self.frame_import, name="button_export")
        self.button_export.configure(text='Export Filters')
        self.button_export.place(anchor="nw", relx=0.6, rely=0.05, x=0, y=0)
        self.button_export.configure(command=self.export_data)


        # max / min SPL control box
        self.separator1 = ttk.Separator(self.frame_import)
        self.separator1.configure(orient="horizontal")
        self.separator1.place(anchor="nw", rely=0.14, width=250, x=0, y=0)
        self.label_minSPL = ttk.Label(self.frame_import, name="label_minspl")
        self.label_minSPL.configure(text='min SPL')
        self.label_minSPL.place(anchor="nw", relx=0.05, rely=0.16, x=0, y=0)
        self.label_maxSPL = ttk.Label(self.frame_import, name="label_maxspl")
        self.label_maxSPL.configure(text='max SPL')
        self.label_maxSPL.place(anchor="nw", relx=0.05, rely=0.2, x=0, y=0)
        self.entry_minSPL = ttk.Entry(self.frame_import, name="entry_minspl")
        self.entry_minSPL.place(anchor="nw", relx=0.25, rely=0.16, width=50, x=0, y=0)
        self.entry_maxSPL = ttk.Entry(self.frame_import, name="entry_maxspl")
        self.entry_maxSPL.place(anchor="nw", relx=0.25, rely=0.20, width=50, x=0, y=0)

        # min / max frequency axis
        self.label_minFreq = ttk.Label(self.frame_import, name='label_minfreq')
        self.label_minFreq.configure(text="min freq")
        self.label_minFreq.place(anchor="nw", relx=0.5, rely=0.16, x=0, y=0)
        self.label_maxFreq = ttk.Label(self.frame_import, name='label_maxfreq')
        self.label_maxFreq.configure(text="max freq")
        self.label_maxFreq.place(anchor="nw", relx=0.5, rely=0.2, x=0, y=0)
        self.entry_minFreq = ttk.Entry(self.frame_import, name="entry_minfreq")
        self.entry_minFreq.place(anchor="nw", relx=0.71, rely=0.16, width=50, x=0, y=0)
        self.entry_maxFreq = ttk.Entry(self.frame_import, name="entry_maxfreq")
        self.entry_maxFreq.place(anchor="nw", relx=0.71, rely=0.20, width=50, x=0, y=0)


        # pack dataset frame
        self.frame_import.pack(side="top")
        self.notebook.add(self.frame_import, text='Dataset')

        # ======================================================================
        # DRIVER TAB DICT - number of tab should be set as the number of drivers
        self.filt_index = ['A', 'B', 'C', 'D', 'E', 'F', 'delay']
        self.rely = [0.05, 0.09, 0.13]
        self.rely_jump = 0.14
        self.DRIVER_FRAME = {}
        self.cb_entry = {}
        self.cb_label = {}
        self.freq_entry = {}
        self.freq_label = {}
        self.q_entry = {}
        self.q_label = {}
        self.gain_entry = {}
        self.gain_label = {}
        self.delay_entry = {}
        self.delay_label = {}
        self.delay_label_ms = {}
        self.ggain_entry = {} # general gain
        self.ggain_label = {} # general gain
        self.ggain_label_db = {} # general gain
        self.flipPhase_box = {}
        self.flipPhase_var = {}
        # self.flipPhase_label = {}
        self.update_button = {}
        self.filter_name = {}
        self.filter_label = {}

        # help with the phase flipping
        self.wasFlipped = {}

        # =================================
        # PLOTTING FRAME
        self.notebook.grid(column=0, row=0)
        self.frame_plot = ttk.Frame(self.panedWindow, name="frame_plot")
        self.frame_plot.configure(height=720, width=1000)
        self.frame_plot.grid(column=1, row=0)

        self.notebook_plot = ttk.Notebook(self.frame_plot,
                                          name="notebook_plot")  # each tab will be a notebook_obs entry
        self.notebook_plot.configure(height=720, width=1000)
        self.notebook_obs = {}      # each tab will be a specific driver
        self.PLOT_FRAME = {}        # to store plot frame in sub-notebooks (for each observation)

        # panedwindow setup
        self.panedWindow.add(self.notebook, weight=1)
        self.panedWindow.add(self.frame_plot, weight=1)
        self.panedWindow.grid(column=0, row=1)

        # plot data (for each observation)
        self.theta = {}
        self.pMic_on_axis = {}
        self.pMic_30_deg = {}

        # ==================
        # finalize app frame
        # pack main frame
        self.frame1.pack()

        # Main widget
        self.mainwindow = self.toplevel1

    def run(self):
        self.mainwindow.mainloop()

    def add_driver_tab(self, index):
        str_ind = str(index)

        # ========================================================================================
        #
        #   ADD DATASET TAB
        #
        # ========================================================================================
        self.DRIVER_FRAME[str_ind] = ttk.Frame(self.notebook)
        self.DRIVER_FRAME[str_ind] = ttk.Frame(self.notebook, name="frame_DRV_{}".format(str_ind))
        self.DRIVER_FRAME[str_ind].configure(height=200, width=200)
        self.DRIVER_FRAME[str_ind].pack(side="top")
        self.notebook.add(self.DRIVER_FRAME[str_ind], text="Drv " + str_ind)

        # =======================
        # CREATE FILTER ENTRIES
        y_offset = 0
        def_freq_values = ["100", "200", "400", "800", "1600", "3200"]
        def_q_value = "0.707"
        def_g_value = "0"
        def_delay_val = "0"
        def_filter_name = "DRV_{}".format(str_ind)
        for i in range(6):
            abc = self.filt_index[i]
            filt = str_ind + abc
            # print('addtab ', filt)
            # COMBOBOX
            self.cb_entry[filt] = ttk.Combobox(self.DRIVER_FRAME[str_ind], name="cb_entry_{}".format(filt))
            self.cb_entry[filt].place(anchor="nw", relx=0.25, rely=self.rely[0] + y_offset, x=0, y=0)
            self.cb_entry[filt]['values'] = ("Lowpass", "Highpass", "Bandpass", "Lowshelf", "Highshelf", "Peaking")
            self.cb_entry[filt].set("Peaking")
            self.cb_label[filt] = ttk.Label(self.DRIVER_FRAME[str_ind], name="cb_label_{}".format(filt))
            self.cb_label[filt].configure(text='Filter {}'.format(abc))
            self.cb_label[filt].place(anchor="nw", relx=0.05, rely=self.rely[0] + y_offset, x=0, y=0)

            # FREQ ENTRY
            self.freq_entry[filt] = ttk.Entry(self.DRIVER_FRAME[str_ind], name="freq_entry_{}".format(filt))
            self.freq_entry[filt].place(anchor="nw", relx=0.25, rely=self.rely[1] + y_offset, width=50, x=0, y=0)
            self.freq_entry[filt].insert(0, def_freq_values[i])
            self.freq_label[filt] = ttk.Label(self.DRIVER_FRAME[str_ind], name="freq_label_{}".format(filt))
            self.freq_label[filt].configure(text='Freq')
            self.freq_label[filt].place(anchor="nw", relx=0.05, rely=self.rely[1] + y_offset, x=0, y=0)

            # Q ENTRY
            self.q_entry[filt] = ttk.Entry(self.DRIVER_FRAME[str_ind], name="q_entry_{}".format(filt))
            self.q_entry[filt].place(anchor="nw", relx=0.25, rely=self.rely[2] + y_offset, width=50, x=0, y=0)
            self.q_entry[filt].insert(0, def_q_value)
            self.q_label[filt] = ttk.Label(self.DRIVER_FRAME[str_ind], name="q_label_{}".format(filt))
            self.q_label[filt].configure(text='Q')
            self.q_label[filt].place(anchor="nw", relx=0.05, rely=self.rely[2] + y_offset, x=0, y=0)

            # GAIN ENTRY
            self.gain_entry[filt] = ttk.Entry(self.DRIVER_FRAME[str_ind], name="g_entry_{}".format(filt))
            self.gain_entry[filt].place(anchor="nw", relx=0.70, rely=self.rely[2] + y_offset, width=50, x=0, y=0)
            self.gain_entry[filt].insert(0, def_g_value)
            self.gain_label[filt] = ttk.Label(self.DRIVER_FRAME[str_ind], name='g_label_{}'.format(filt))
            self.gain_label[filt].configure(text='dB')
            self.gain_label[filt].place(anchor="nw", relx=0.6, rely=self.rely[2] + y_offset, x=0, y=0)

            y_offset += self.rely_jump
            # exit loop

        # delay
        self.delay_label[str_ind] = ttk.Label(self.DRIVER_FRAME[str_ind], name="delay_label_{}".format(str_ind))
        self.delay_label[str_ind].configure(text='Delay')
        self.delay_label[str_ind].place(anchor="nw", relx=0.05, rely=0.88, width=50, x=0, y=0)
        self.delay_entry[str_ind] = ttk.Entry(self.DRIVER_FRAME[str_ind], name="delay_entry_{}".format(str_ind))
        self.delay_entry[str_ind].place(anchor='nw', relx=0.25, rely=0.88, width=50, x=0, y=0)
        self.delay_entry[str_ind].insert(0, def_delay_val)
        self.delay_label_ms[str_ind] = ttk.Label(self.DRIVER_FRAME[str_ind], name="delay_ms_{}".format(str_ind))
        self.delay_label_ms[str_ind].configure(text='(ms)')
        self.delay_label_ms[str_ind].place(anchor="nw", relx=0.46, rely=0.88, x=0, y=0)

        # gain
        self.ggain_label[str_ind] = ttk.Label(self.DRIVER_FRAME[str_ind], name="ggain_label_{}".format(str_ind))
        self.ggain_label[str_ind].configure(text='Gain')
        self.ggain_label[str_ind].place(anchor="nw", relx=0.05, rely=0.92, width=50, x=0, y=0)
        self.ggain_entry[str_ind] = ttk.Entry(self.DRIVER_FRAME[str_ind], name="ggain_entry_{}".format(str_ind))
        self.ggain_entry[str_ind].place(anchor='nw', relx=0.25, rely=0.92, width=50, x=0, y=0)
        self.ggain_entry[str_ind].insert(0, def_delay_val)
        self.ggain_label_db[str_ind] = ttk.Label(self.DRIVER_FRAME[str_ind], name="ggain_dB_{}".format(str_ind))
        self.ggain_label_db[str_ind].configure(text='(dB)')
        self.ggain_label_db[str_ind].place(anchor="nw", relx=0.46, rely=0.92, x=0, y=0)

        # flip phase
        # self.flipPhase_var[str_ind] = tk.BooleanVar()
        self.flipPhase_box[str_ind] = ttk.Checkbutton(self.DRIVER_FRAME[str_ind],
                                                      name="flip_checkbox_{}".format(str_ind))
        self.flipPhase_box[str_ind].configure(text='Flip phase')
        self.flipPhase_box[str_ind].place(anchor="nw", relx=0.05, rely=0.96, x=0, y=0)
        self.flipPhase_box[str_ind].state(['!alternate'])

        # update button
        self.update_button[str_ind] = ttk.Button(self.DRIVER_FRAME[str_ind], name="update_filt_{}".format(str_ind),
                                                 command=self.update_xovers)
        self.update_button[str_ind].configure(text='Update Filter')
        self.update_button[str_ind].place(anchor="nw", relx=0.65, rely=0.95, x=0, y=0)

        # filter / driver name
        self.filter_name[str_ind] = ttk.Entry(self.DRIVER_FRAME[str_ind], name='filter_name_{}'.format(str_ind))
        self.filter_name[str_ind].place(anchor="nw", relx=0.25, rely=0.01, x=0, y=0)
        self.filter_name[str_ind].insert(0, def_filter_name)
        self.filter_label[str_ind] = ttk.Label(self.DRIVER_FRAME[str_ind], name='filter_label_{}'.format(str_ind))
        self.filter_label[str_ind].configure(text='Name')
        self.filter_label[str_ind].place(anchor="nw", relx=0.05, rely=0.01, x=0, y=0)
        return None

    def add_observation_tab(self, obsName, indexDriver):
        str_ind = obsName + "_" + str(indexDriver)
        text_str = "DRV_" + str(indexDriver)
        # ========================================================================================
        #
        #   ADD PLOTTING TAB
        #
        # ========================================================================================
        self.PLOT_FRAME[str_ind] = ttk.Frame(self.notebook_obs[obsName], name="frame_plot_{}".format(str_ind))
        self.PLOT_FRAME[str_ind].configure(height=200, width=200)
        self.PLOT_FRAME[str_ind].pack(side="top")
        self.notebook_obs[obsName].add(self.PLOT_FRAME[str_ind], text=text_str)  # self.filter_name[str_ind].get()

        # plot data
        # dataset = self.pMic[obsName]
        # print("add_pressure ", str_ind)
        self.plot_data(self.PLOT_FRAME[str_ind], obsName, indexDriver - 1, 0)
        return None

    def add_summed_pressure_tab(self, obsName):
        str_ind = obsName + "_all"
        text_str = "DRV_" + "_all"

        self.PLOT_FRAME[str_ind] = ttk.Frame(self.notebook_obs[obsName], name="frame_plot_{}".format(str_ind))
        self.PLOT_FRAME[str_ind].configure(height=200, width=200)
        self.PLOT_FRAME[str_ind].pack(side="top")
        self.notebook_obs[obsName].add(self.PLOT_FRAME[str_ind], text=text_str)

        self.plot_data_summed(self.PLOT_FRAME[str_ind], obsName, 0)
        return None

    def load_data(self):
        self.data_path = filedialog.askdirectory()  # file should be an observation npz archive
        dat = load(self.data_path)
        self.dat = dat

        for study in dat.acoustic_study:
            self.frequency = dat.acoustic_study[study].freq_array
            for obs in range(len(dat.observation[study].observationName)):
                if dat.observation[study].observationType[obs] == 'polar':
                    obsName = dat.observation[study].observationName[obs]
                    updateResults(dat, study, bypass_xover=True)
                    self.observation[obsName] = obsName
                    self.pMic[obsName] = dat.results[study].pMicArray[obs]
                    self.angle[obsName] = dat.observation[study].theta[obs]

        self.study = study
        self.entry_minFreq.insert(0, str(int(self.frequency[0])))
        self.entry_maxFreq.insert(0, str(int(self.frequency[-1])))
        iteration = 0
        for obs in self.observation:
            # initialize observation + plotting frames
            self.notebook_obs[obs] = ttk.Notebook(self.notebook_plot, name="notebook_obs_{}".format(obs))
            self.notebook_obs[obs].configure(width=200, height=200)

            shape_pMic = self.pMic[obs].shape
            nDriver = shape_pMic[-1]
            for i in range(nDriver):
                if iteration == 0:
                    self.xovers["drv_" + str(i + 1)] = xovers(self.frequency)
                    self.wasFlipped["drv_" + str(i+1)] = False
                    self.add_driver_tab(i + 1)
                self.add_observation_tab(obs, i + 1)
            iteration += 1
            self.add_summed_pressure_tab(obs)
            self.notebook_obs[obs].pack(side="top")
            self.notebook_plot.add(self.notebook_obs[obs], text="{}".format(obs))
        self.notebook_plot.pack(side="top")
        return None

    def export_data(self):
        for i, xo in enumerate(self.xovers):
            self.dat.crossover[xo] = self.xovers[xo]
            self.dat.crossover[xo].referenceStudy = self.study
            # print("here {}".format(xo))
            # for i in range(len(self.dat.acoustic_study[self.study].radiatingSurface)):
            self.dat.crossover[xo].ref2bem = int(self.dat.acoustic_study[self.study].radiatingSurface[i])
                # print("filter {} exported as ref to {}".format(xo, int(self.dat.acoustic_study[self.study].radiatingSurface[i])))
        save(self.data_path+"_xoverDesigner", self.dat)
        return None


    def plot_data(self, masterWindow, obs, index_driver, filters):
        # import matplotlib.pyplot as plt
        ## process dataset and store initial data
        theta = self.angle[obs] * 180 / np.pi
        idx_z, _ = gtb.findInArray(self.angle[obs], 0)
        idx_30, _ = gtb.findInArray(theta, 30)
        N_angle = len(self.angle[obs])
        h = self.xovers["drv_" + str(index_driver + 1)].h

        pMic_XO = self.pMic[obs][:, :, index_driver]

        # on-axis SPL
        SPL_on_axis = gtb.gain.SPL(pMic_XO[:, idx_z])
        SPL_30_deg = gtb.gain.SPL(pMic_XO[:, idx_30])

        # SPL directivity + get default min-max SPL values
        SPL_directivity = gtb.gain.SPL(pMic_XO)
        MAX_SPL = int(np.max(SPL_directivity) + 3)
        MIN_SPL = MAX_SPL - 40
        try:
            if int(self.entry_minSPL.get()) > MIN_SPL:
                self.entry_minSPL.delete(0, "end")
                self.entry_minSPL.insert(0, str(MIN_SPL))
            if int(self.entry_maxSPL.get()) < MAX_SPL:
                self.entry_maxSPL.delete(0, "end")
                self.entry_maxSPL.insert(0, str(MAX_SPL))
        except:
            self.entry_minSPL.insert(0, str(MIN_SPL))
            self.entry_maxSPL.insert(0, str(MAX_SPL))

        # absolute directivity
        maxAngle = np.array([np.max(abs(pMic_XO), 1)])
        maxMatrix = np.repeat(maxAngle, len(self.angle[obs]), 0).T
        ABS_directivity = gtb.gain.dB(np.abs(pMic_XO) / maxMatrix)

        # min / max frequency bounds
        minf = float(self.entry_minFreq.get())
        maxf = float(self.entry_maxFreq.get())

        # plot fig
        indexFig = obs+str(index_driver + 1)
        self.fig[indexFig] = Figure(figsize=(10.417, 7.5))
        # plot 1
        plot1 = self.fig[indexFig].add_subplot(221)
        gca = plot1.contourf(self.frequency, theta, SPL_directivity.T, np.arange(MIN_SPL, MAX_SPL, 3),
                             cmap="turbo")
        plot1.set_xscale('log')
        cbar = self.fig[indexFig].colorbar(gca)
        cbar.set_label('SPL [dB]')
        plot1.set_xlabel("Frequency [Hz]")
        plot1.set_ylabel("Angle [degree]")
        plot1.set_xlim([minf, maxf])

        # plot 2
        plot2 = self.fig[indexFig].add_subplot(222)
        plot2.semilogx(self.frequency, SPL_on_axis, 'b', label='on-axis')
        plot2.semilogx(self.frequency, SPL_30_deg, 'g', label='30 deg', alpha=0.75, linewidth=0.7)
        plot2.set_xlabel('Frequency [Hz]')
        plot2.set_ylabel('SPL [dB]')
        plot2.grid(True, linestyle='dotted', which='both')
        plot2.legend(loc='best')
        plot2.set_xlim([minf, maxf])
        plot2.set_ylim([MIN_SPL, MAX_SPL])

        # plot3
        plot3 = self.fig[indexFig].add_subplot(223)
        gca3 = plot3.contourf(self.frequency, theta, ABS_directivity.T,
                              np.arange(-21, 3, 3), cmap="turbo")
        plot3.set_xscale('log')
        cbar3 = self.fig[indexFig].colorbar(gca3)
        cbar3.set_label('Gain [dB]')
        plot3.set_xlabel("Frequency [Hz]")
        plot3.set_ylabel("Angle [degree]")
        plot3.set_xlim([minf, maxf])

        # plot 4
        plot4 = self.fig[indexFig].add_subplot(224)
        plot4.semilogx(self.frequency, gtb.gain.dB(h), 'k')
        plot4.set_xlabel('Frequency [Hz]')
        plot4.set_ylabel('Gain [dB]')
        # plot4.set_ylim([-32, 3])
        plot4.grid(True, linestyle='dotted', which='both')
        plot4.set_ylim([-32, +32])

        plot4_p = plot4.twinx()
        plot4_p.semilogx(self.frequency, np.angle(h), ':r')
        plot4_p.set_ylim([-3.2, 3.2])
        plot4_p.set_ylabel("Phase [rad]", color='red')
        plot4_p.tick_params(axis='y', labelcolor='red')

        self.fig[indexFig].tight_layout()
        self.canvas[indexFig] = FigureCanvasTkAgg(self.fig[indexFig], master=masterWindow)
        self.canvas[indexFig].draw()

        # placing the canvas on the Tkinter window
        self.canvas[indexFig].get_tk_widget().pack()
        # print("index Fig - plot data: ", indexFig)
        # self.canvas_widget[indexFig] = canvas.get_tk_widget()
        # self.canvas_widget[indexFig].pack()
        # toolbar = NavigationToolbar2Tk(self.canvas[indexFig], masterWindow)
        # toolbar.update()

        # placing the toolbar on the Tkinter window
        # self.canvas_widget[indexFig].get_tk_widget().pack()
        return None

    def update_plot_data(self, masterWindow, obs, index_driver, filters):
        ## process dataset
        theta = self.angle[obs] * 180 / np.pi
        idx_z, _ = gtb.findInArray(self.angle[obs], 0)
        idx_30, _ = gtb.findInArray(theta, 30)
        N_angle = len(self.angle[obs])
        h = self.xovers["drv_" + str(index_driver + 1)].h

        XO = np.zeros([len(self.frequency), N_angle], dtype=complex)
        for i in range(N_angle):
            XO[:, i] = h
        pMic_XO = self.pMic[obs][:, :, index_driver] * XO

        # on-axis SPL
        SPL_on_axis = gtb.gain.SPL(pMic_XO[:, idx_z])
        SPL_30_deg = gtb.gain.SPL(pMic_XO[:, idx_30])

        # SPL directivity + set min/max SPL
        SPL_directivity = gtb.gain.SPL(pMic_XO)
        MIN_SPL = int(self.entry_minSPL.get())
        MAX_SPL = int(self.entry_maxSPL.get())

        # absolute directivity
        maxAngle = np.array([np.max(abs(pMic_XO), 1)])
        maxMatrix = np.repeat(maxAngle, len(self.angle[obs]), 0).T
        ABS_directivity = gtb.gain.dB(np.abs(pMic_XO) / maxMatrix)

        # min / max frequency bounds
        minf = float(self.entry_minFreq.get())
        maxf = float(self.entry_maxFreq.get())

        # plot fig
        indexFig = obs + str(index_driver + 1)
        self.fig[indexFig].clear()
        # plot 1
        plot1 = self.fig[indexFig].add_subplot(221)
        gca = plot1.contourf(self.frequency, theta, SPL_directivity.T, np.arange(MIN_SPL, MAX_SPL, 3),
                                            cmap="turbo")
        plot1.set_xscale('log')
        cbar = self.fig[indexFig].colorbar(gca)
        cbar.set_label('SPL [dB]')
        plot1.set_xlabel("Frequency [Hz]")
        plot1.set_ylabel("Angle [degree]")
        plot1.set_xlim([minf, maxf])

        # plot 2
        plot2 = self.fig[indexFig].add_subplot(222)
        plot2.semilogx(self.frequency, SPL_on_axis, 'b', label='on-axis')
        plot2.semilogx(self.frequency, SPL_30_deg, 'g', label='30 deg', alpha=0.75, linewidth=0.7)
        plot2.set_xlabel('Frequency [Hz]')
        plot2.set_ylabel('SPL [dB]')
        plot2.grid(True, linestyle='dotted', which='both')
        plot2.set_xlim([minf, maxf])
        plot2.set_ylim([MIN_SPL, MAX_SPL])
        plot2.legend(loc='best')

        # plot3
        plot3 = self.fig[indexFig].add_subplot(223)
        gca3 = plot3.contourf(self.frequency, theta, ABS_directivity.T,
                                             np.arange(-21, 3, 3), cmap="turbo")
        plot3.set_xscale('log')
        cbar3 = self.fig[indexFig].colorbar(gca3)
        cbar3.set_label('Gain [dB]')
        plot3.set_xlabel("Frequency [Hz]")
        plot3.set_ylabel("Angle [degree]")
        plot3.set_xlim([minf, maxf])

        # plot 4
        plot4 = self.fig[indexFig].add_subplot(224)
        plot4.semilogx(self.frequency, gtb.gain.dB(h), 'k')
        plot4.set_xlabel('Frequency [Hz]')
        plot4.set_ylabel('Gain [dB]')
        plot4.set_ylim([-32, +32])
        plot4.grid(True, linestyle='dotted', which='both')

        plot4_p = plot4.twinx()
        plot4_p.semilogx(self.frequency, np.angle(h), ':r')
        plot4_p.set_ylim([-3.2, 3.2])
        plot4_p.set_ylabel("Phase [rad]", color='red')
        plot4_p.tick_params(axis='y', labelcolor='red')

        self.canvas[indexFig].draw()
        return None

    def plot_data_summed(self, masterWindow, obs, filter):
        ## process dataset
        theta = self.angle[obs] * 180 / np.pi
        idx_z, _ = gtb.findInArray(self.angle[obs], 0)
        idx_30, _ = gtb.findInArray(theta, 30)
        N_angle = len(self.angle[obs])
        N_driver = np.shape(self.pMic[obs])[-1]
        pMic_XO = np.zeros([len(self.frequency), N_angle], dtype=complex)
        for i in range(N_driver):
            h = self.xovers["drv_" + str(i + 1)].h
            XO = np.tile(h, (N_angle, 1))
            pMic_XO += self.pMic[obs][:, :, i] * XO.T

        # on-axis SPL
        SPL_on_axis = gtb.gain.SPL(pMic_XO[:, idx_z])
        SPL_30_deg = gtb.gain.SPL(pMic_XO[:, idx_30])

        # SPL directivity + set min/max SPL
        SPL_directivity = gtb.gain.SPL(pMic_XO)
        MAX_SPL = int(np.max(SPL_directivity) + 3)
        MIN_SPL = MAX_SPL - 40

        try:
            if int(self.entry_minSPL.get()) > MIN_SPL:
                self.entry_minSPL.delete(0, "end")
                self.entry_minSPL.insert(0, str(MIN_SPL))
            if int(self.entry_maxSPL.get()) < MAX_SPL:
                self.entry_maxSPL.delete(0, "end")
                self.entry_maxSPL.insert(0, str(MAX_SPL))
        except:
            self.entry_minSPL.insert(0, str(MIN_SPL))
            self.entry_maxSPL.insert(0, str(MAX_SPL))

        # absolute directivity
        maxAngle = np.array([np.max(abs(pMic_XO), 1)])
        maxMatrix = np.repeat(maxAngle, len(self.angle[obs]), 0).T
        ABS_directivity = gtb.gain.dB(np.abs(pMic_XO) / maxMatrix)

        # min / max frequency bounds
        minf = float(self.entry_minFreq.get())
        maxf = float(self.entry_maxFreq.get())

        #
        index_fig = obs + "_all"
        self.fig_sum[index_fig] = Figure(figsize=(10.417, 7.5))
        # plot 1
        plot1 = self.fig_sum[index_fig].add_subplot(221)
        gca = plot1.contourf(self.frequency, theta, SPL_directivity.T, np.arange(MIN_SPL, MAX_SPL, 3),
                             cmap="turbo")
        plot1.set_xscale('log')
        cbar = self.fig_sum[index_fig].colorbar(gca)
        cbar.set_label('SPL [dB]')
        plot1.set_xlabel("Frequency [Hz]")
        plot1.set_ylabel("Angle [degree]")
        plot1.set_xlim([minf, maxf])

        # plot 2
        plot2 = self.fig_sum[index_fig].add_subplot(222)
        plot2.semilogx(self.frequency, SPL_on_axis, 'b', label="on-axis")
        plot2.semilogx(self.frequency, SPL_30_deg, 'g', label="30 deg", alpha=0.75, linewidth=0.7)
        plot2.set_xlabel('Frequency [Hz]')
        plot2.set_ylabel('SPL [dB]')
        plot2.set_xlim([minf, maxf])
        plot2.set_ylim([MIN_SPL, MAX_SPL])
        plot2.grid(True, linestyle='dotted', which='both')
        plot2.legend(loc='best')

        # plot3
        plot3 = self.fig_sum[index_fig].add_subplot(223)
        gca3 = plot3.contourf(self.frequency, theta, ABS_directivity.T,
                              np.arange(-21, 3, 3), cmap="turbo")
        plot3.set_xscale('log')
        cbar3 = self.fig_sum[index_fig].colorbar(gca3)
        cbar3.set_label('Gain [dB]')
        plot3.set_xlabel("Frequency [Hz]")
        plot3.set_ylabel("Angle [degree]")
        plot3.set_xlim([minf, maxf])

        # plot 4
        plot4 = self.fig_sum[index_fig].add_subplot(224)
        plot4.semilogx(self.frequency, gtb.gain.dB(h), 'k')
        plot4.set_xlabel('Frequency [Hz]')
        plot4.set_ylabel('Gain [dB]')
        plot4.set_ylim([-32, 32])
        plot4.grid(True, linestyle='dotted', which='both')

        plot4_p = plot4.twinx()
        plot4_p.semilogx(self.frequency, np.angle(h), ':r')
        plot4_p.set_ylim([-3.2, 3.2])
        plot4_p.set_ylabel("Phase [rad]", color='red')
        plot4_p.tick_params(axis='y', labelcolor='red')

        self.fig_sum[index_fig].tight_layout()
        self.canvas_sum[index_fig] = FigureCanvasTkAgg(self.fig_sum[index_fig], master=masterWindow)
        self.canvas_sum[index_fig].draw()

        # placing the canvas on the Tkinter window
        self.canvas_sum[index_fig].get_tk_widget().pack()
        return None


    def update_plot_data_summed(self, masterWindow, obs, filter):
        ## process dataset
        theta = self.angle[obs] * 180 / np.pi
        idx_z, _ = gtb.findInArray(self.angle[obs], 0)
        idx_30, _ = gtb.findInArray(theta, 30)
        N_angle = len(self.angle[obs])
        N_driver = np.shape(self.pMic[obs])[-1]
        pMic_XO = np.zeros([len(self.frequency), N_angle], dtype=complex)
        h_to_plot = np.zeros(len(self.frequency), dtype=complex)
        for i in range(N_driver):
            h = self.xovers["drv_" + str(i + 1)].h
            h_to_plot += h
            XO = np.tile(h, (N_angle, 1))
            pMic_XO += self.pMic[obs][:, :, i] * XO.T

        # on-axis SPL
        SPL_on_axis = gtb.gain.SPL(pMic_XO[:, idx_z])
        SPL_30_deg = gtb.gain.SPL(pMic_XO[:, idx_30])

        # SPL directivity
        SPL_directivity = gtb.gain.SPL(pMic_XO)
        MIN_SPL = int(self.entry_minSPL.get())
        MAX_SPL = int(self.entry_maxSPL.get())
        # MAX_SPL = int(np.max(SPL_directivity) + 3)
        # MIN_SPL = MAX_SPL - 40


        # absolute directivity
        maxAngle = np.array([np.max(abs(pMic_XO), 1)])
        maxMatrix = np.repeat(maxAngle, len(self.angle[obs]), 0).T
        ABS_directivity = gtb.gain.dB(np.abs(pMic_XO) / maxMatrix)

        # min / max frequency bounds
        minf = float(self.entry_minFreq.get())
        maxf = float(self.entry_maxFreq.get())

        index_fig = obs + "_all"
        self.fig_sum[index_fig].clear()
        # plot 1
        plot1 = self.fig_sum[index_fig].add_subplot(221)
        gca = plot1.contourf(self.frequency, theta, SPL_directivity.T, np.arange(MIN_SPL, MAX_SPL, 3),
                             cmap="turbo")
        plot1.set_xscale('log')
        cbar = self.fig_sum[index_fig].colorbar(gca)
        cbar.set_label('SPL [dB]')
        plot1.set_xlabel("Frequency [Hz]")
        plot1.set_ylabel("Angle [degree]")
        plot1.set_xlim([minf, maxf])

        # plot 2
        plot2 = self.fig_sum[index_fig].add_subplot(222)
        plot2.semilogx(self.frequency, SPL_on_axis, 'b', label='on-axis')
        plot2.semilogx(self.frequency, SPL_30_deg, 'g', label='30 deg', alpha=0.75, linewidth=0.7)
        plot2.set_xlabel('Frequency [Hz]')
        plot2.set_ylabel('SPL [dB]')
        plot2.grid(True, linestyle='dotted', which='both')
        plot2.set_xlim([minf, maxf])
        plot2.set_ylim([MIN_SPL, MAX_SPL])
        plot2.legend(loc='best')

        # plot3
        plot3 = self.fig_sum[index_fig].add_subplot(223)
        gca3 = plot3.contourf(self.frequency, theta, ABS_directivity.T,
                              np.arange(-21, 3, 3), cmap="turbo")
        plot3.set_xscale('log')
        cbar3 = self.fig_sum[index_fig].colorbar(gca3)
        cbar3.set_label('Gain [dB]')
        plot3.set_xlabel("Frequency [Hz]")
        plot3.set_ylabel("Angle [degree]")
        plot3.set_xlim([minf, maxf])

        # plot 4
        plot4 = self.fig_sum[index_fig].add_subplot(224)
        plot4.semilogx(self.frequency, gtb.gain.dB(h_to_plot), 'k')
        plot4.set_xlabel('Frequency [Hz]')
        plot4.set_ylabel('Gain [dB]')
        plot4.set_ylim([-32, 32])
        plot4.grid(True, linestyle='dotted', which='both')

        plot4_p = plot4.twinx()
        plot4_p.semilogx(self.frequency, np.angle(h_to_plot), ':r')
        plot4_p.set_ylim([-3.2, 3.2])
        plot4_p.set_ylabel("Phase [rad]", color='red')
        plot4_p.tick_params(axis='y', labelcolor='red')

        self.canvas_sum[index_fig].draw()
        return None


    def update_xovers(self):
        iteration = 0
        for obs in self.observation:
            N_driver = np.shape(self.pMic[obs])[-1]
            for n in range(N_driver):
                for i in range(6):
                    filt_index = str(n+1) + self.filt_index[i]
                    # print('xo: ', filt_index)
                    filtertype = self.cb_entry[filt_index].get()
                    fc = float(self.freq_entry[filt_index].get())
                    Q = float(self.q_entry[filt_index].get())
                    dB = float(self.gain_entry[filt_index].get())
                    self.assign_xover(filtertype, fc, Q, dB, str(n + 1), filt_index)
                delay = float(self.delay_entry[str(n+1)].get())
                flipPhase = self.flipPhase_box[str(n+1)].state()
                gdB = float(self.ggain_entry[str(n+1)].get()) # general gain
                self.assign_gain_delay_flip(gdB, delay, flipPhase, str(n+1))
                str_ind = obs + "_" + str(n+1)
                self.update_plot_data(self.PLOT_FRAME[str_ind], obs, n, 0)
            self.update_plot_data_summed(self.PLOT_FRAME[obs+"_all"], obs, 0)
        return None

    def assign_xover(self, filtertype, fc, Q, dB, spk_index, filt_index):
        str_drv = "drv_" + spk_index
        if filtertype == 'Lowpass':
            # print('lowpass')
            self.xovers[str_drv].addLowPassBQ(filt_index, fc, Q, dB)
        elif filtertype == 'Highpass':
            # print('highpass')
            self.xovers[str_drv].addHighPassBQ(filt_index, fc, Q, dB)
        elif filtertype == 'Bandpass':
            # print('bandpass')
            self.xovers[str_drv].addBandPassBQ(filt_index, fc, Q, dB)
        elif filtertype == 'Peaking':
            # print('peaking')
            self.xovers[str_drv].addPeakEQ(filt_index, fc, Q, dB)
        elif filtertype == 'Lowshelf':
            # print('lowshelf')
            self.xovers[str_drv].addLowShelf(filt_index, fc, Q, dB)
        elif filtertype == 'Highshelf':
            # print('highshelf')
            self.xovers[str_drv].addHighShelf(filt_index, fc, Q, dB)
        else:
            print("filter not understood")
        return None

    def assign_gain_delay_flip(self, gain, delay, isFlipped, spk_index): # could be better
        str_drv = "drv_" + spk_index
        self.xovers[str_drv].addGain('dB', gain)
        self.xovers[str_drv].addDelay('dt', delay*1e-3)
        try:
            isFlipped = isFlipped[0]
        except:
            pass
        if isFlipped == 'selected':
            if self.wasFlipped[str_drv] is False:
                self.xovers[str_drv].addPhaseFlip('pi')
                self.wasFlipped[str_drv] = True
            elif self.wasFlipped[str_drv] is True:
                pass
        elif len(isFlipped) == 0:
            if self.wasFlipped[str_drv] is True:
                self.xovers[str_drv].deleteFilter('pi')
                self.wasFlipped[str_drv] = False
            elif self.wasFlipped[str_drv] is False:
                pass
        return None

if __name__ == "__main__":
    app = XoverAppApp()
    app.run()



"""Data structures for assembled boundary operators."""

import warnings
import numpy as _np
from bempp_cl.helpers import timeit as _timeit
# from scipy.sparse.linalg.interface import LinearOperator as _LinearOperator
from scipy.sparse.linalg import LinearOperator as _LinearOperator

# Disable warnings for differing overridden parameters
# pylint: disable=W0221


class _DiscreteOperatorBase(_LinearOperator):
    """Discrete boundary operator base."""

    def __init__(self, dtype, shape):
        """Construct. Should not be called directly."""
        super().__init__(dtype, shape)

    def __add__(self, other):
        """Sum of two operators."""
        if isinstance(other, _DiscreteOperatorBase):
            return _SumDiscreteOperator(self, other)
        else:
            return super().__add__(other)

    def __neg__(self):
        """Negation."""
        return _ScaledDiscreteOperator(self, -1)

    def __sub__(self, other):
        """Subtraction."""
        return self.__add__(-other)

    def dot(self, other):
        """Product with other objects."""
        if isinstance(other, _DiscreteOperatorBase):
            return _ProductDiscreteOperator(self, other)
        elif _np.isscalar(other):
            return _ScaledDiscreteOperator(self, other)
        else:
            return super().dot(other)

    def __mul__(self, other):
        """Product with other objects."""
        return self.dot(other)

    def __rmul__(self, other):
        """Reverse product."""
        if _np.isscalar(other):
            return _ScaledDiscreteOperator(self, other)
        else:
            return NotImplemented

    @property
    def A(self):
        """Return dense matrix."""
        warnings.warn(
            "operator.A is deprecated and will be removed in a future version. Use operator.to_dense() instead.",
            DeprecationWarning,
        )
        return self.to_dense()

    def to_dense(self):
        """Return dense matrix."""
        raise NotImplementedError()

    def to_sparse(self):
        """Return sparse matrix if operator is sparse."""
        raise NotImplementedError()


class _ScaledDiscreteOperator(_DiscreteOperatorBase):
    """Return a scaled operator."""

    def __init__(self, op, alpha):
        dtype = _np.result_type(op.dtype, type(alpha))
        self._op = op
        self._alpha = alpha
        super().__init__(dtype, op.shape)

    def _matvec(self, x):
        """Matvec."""
        return self._alpha * (self._op @ x)

    def to_dense(self):
        """Return dense matrix."""
        return self._alpha * self._op.to_dense()

    def to_sparse(self):
        """Return sparse matrix if operator is sparse."""
        return self._alpha * self._op.to_sparse()


class _SumDiscreteOperator(_DiscreteOperatorBase):
    """Return a sum operator."""

    def __init__(self, op1, op2):
        """Construct."""
        if op1.shape != op2.shape:
            raise ValueError(f"Operators have incompatible shapes {op1.shape} != {op2.shape}")

        self._op1 = op1
        self._op2 = op2

        dtype = _np.result_type(op1.dtype, op2.dtype)

        super().__init__(dtype, op1.shape)

    def _matvec(self, x):
        """Evaluate matvec."""
        return self._op1 @ x + self._op2 @ x

    def to_dense(self):
        """Return dense matrix."""
        return self._op1.to_dense() + self._op2.to_dense()

    def to_sparse(self):
        """Return sparse matrix if operator is sparse."""
        return self._op1.to_sparse() + self._op2.to_sparse()


class _ProductDiscreteOperator(_DiscreteOperatorBase):
    """Product of two operators."""

    def __init__(self, op1, op2):
        """Construct the product of two operators."""
        if op1.shape[1] != op2.shape[0]:
            raise ValueError(f"Incompatible dimensions shapes for multiplication with {op1.shape} and {op2.shape}")

        self._op1 = op1
        self._op2 = op2

        dtype = _np.result_type(op1.dtype, op2.dtype)

        super().__init__(dtype, (op1.shape[0], op2.shape[1]))

    def _matvec(self, x):
        """Evaluate matvec."""
        return self._op1 @ (self._op2 @ x)

    def to_dense(self):
        """Return dense matrix."""
        return self._op1.to_dense() @ self._op2.to_dense()

    def to_sparse(self):
        """Return sparse matrix if operator is sparse."""
        return self._op1.to_sparse() @ self._op2.to_sparse()


class GenericDiscreteBoundaryOperator(_DiscreteOperatorBase):
    """Discrete boundary operator that implements a matvec routine."""

    def __init__(self, evaluator):
        """Construct discrete boundary operator."""
        super().__init__(evaluator.dtype, evaluator.shape)
        self._evaluator = evaluator
        self._is_complex = self.dtype == "complex128" or self.dtype == "complex64"

    def _matvec(self, x):
        """Mutliply the operator by a vector."""
        if self._is_complex:
            return self._evaluator.matvec(x)
        if _np.iscomplexobj(x):
            return self._evaluator.matvec(_np.real(x)) + 1j * self._evaluator.matvec(_np.imag(x))
        else:
            return self._evaluator.matvec(x)

    def to_dense(self):
        """Return dense matrix."""
        return self @ _np.eye(self.shape[1])

    def to_sparse(self):
        """Return sparse matrix if operator is sparse."""
        raise ValueError("This matrix is not sparse.")


class DenseDiscreteBoundaryOperator(_DiscreteOperatorBase):
    """
    Main class for the discrete form of dense nonlocal operators.

    This class derives from
    :class:`scipy.sparse.linalg.interface.LinearOperator`
    and thereby implements the SciPy LinearOperator protocol.

    """

    def __init__(self, impl):
        """Construct. Should not be called by the user."""
        self._impl = impl
        super().__init__(impl.dtype, impl.shape)

    def _matmat(self, x):
        """Multiply the operator by a matrix or operator."""
        if _np.iscomplexobj(x) and not _np.iscomplexobj(self.to_dense()):
            return self.to_dense().dot(_np.real(x).astype(self.dtype)) + 1j * self.to_dense().dot(
                _np.imag(x).astype(self.dtype)
            )
        return self.to_dense().dot(x.astype(self.dtype))

    def __add__(self, other):
        """Add two operators."""
        if isinstance(other, DenseDiscreteBoundaryOperator):
            return DenseDiscreteBoundaryOperator(self.to_dense() + other.to_dense())
        else:
            return super().__add__(other)

    def __neg__(self):
        """Negate the operator."""
        return DenseDiscreteBoundaryOperator(-self.to_dense())

    def __mul__(self, other):
        """Multiply."""
        return self.dot(other)

    def dot(self, other):
        """Form the product with another object."""
        if isinstance(other, DenseDiscreteBoundaryOperator):
            return DenseDiscreteBoundaryOperator(self.to_dense().dot(other.to_dense()))
        if _np.isscalar(other):
            if self._impl.dtype in ["float32", "complex64"]:
                # Necessary to ensure that scalar multiplication does not change
                # precision to double precision.
                if _np.iscomplexobj(other):
                    return DenseDiscreteBoundaryOperator(self.to_dense() * _np.dtype("complex64").type(other))
                else:
                    return DenseDiscreteBoundaryOperator(self.to_dense() * _np.dtype("float32").type(other))
            else:
                return DenseDiscreteBoundaryOperator(self.to_dense() * other)
        return super().dot(other)

    def __rmul__(self, other):
        """Multiply."""
        if _np.isscalar(other):
            return DenseDiscreteBoundaryOperator(self.to_dense() * other)
        else:
            return NotImplemented

    def _transpose(self):
        """Transpose of the operator."""
        return DenseDiscreteBoundaryOperator(self.to_dense().T)

    def _adjoint(self):
        """Adjoint of the operator."""
        return DenseDiscreteBoundaryOperator(self.to_dense().conjugate().transpose())

    def to_dense(self):
        """Return dense matrix."""
        return self._impl

    def to_sparse(self):
        """Return sparse matrix if operator is sparse."""
        raise ValueError("This matrix is not sparse.")


class DiagonalOperator(_DiscreteOperatorBase):
    """
    Main class for discrete diagonal operators.

    This class derives from
    :class:`scipy.sparse.linalg.interface.LinearOperator`
    and thereby implements the SciPy LinearOperator protocol.

    """

    def __init__(self, values, shape=None):
        """Construct. Should not be called by the user."""
        self._values = values.ravel()
        if shape is None:
            shape = (len(values), len(values))
        super().__init__(values.dtype, shape)

    def get_diagonal(self):
        """Get the diagonal values."""
        return self._values

    def _matvec(self, x):
        """Multiply the operator by a vector."""
        vec_shape = x.shape

        return (self.get_diagonal() * x.ravel()).reshape(vec_shape)

    def __add__(self, other):
        """Add."""
        if self.shape != other.shape:
            raise ValueError(f"Incompatible dimensions: {self.shape} != {other.shape}")
        if isinstance(other, DiagonalOperator):
            return DiagonalOperator(self.get_diagonal() + other.get_diagonal())
        else:
            return super().__add__(other)

    def __neg__(self):
        """Negate."""
        return DiagonalOperator(-self.get_diagonal())

    def __mul__(self, other):
        """Multiply."""
        return self.dot(other)

    def dot(self, other):
        """Product with other objects."""
        if _np.isscalar(other):
            if self.get_diagonal().dtype in ["float32", "complex64"]:
                # Necessary to ensure that scalar multiplication does not change
                # precision to double precision.
                if _np.iscomplexobj(other):
                    return DiagonalOperator(self.get_diagonal() * _np.dtype("complex64").type(other))
                else:
                    return DiagonalOperator(self.get_diagonal() * _np.dtype("float32").type(other))
            else:
                return DiagonalOperator(self.get_diagonal() * other)
        return super().dot(other)

    def __rmul__(self, other):
        """Multiply."""
        if _np.isscalar(other):
            return DiagonalOperator(self.get_diagonal() * other)
        else:
            return NotImplemented

    def _transpose(self):
        """Transpose of the operator."""
        return self

    def _adjoint(self):
        """Adjoint of the operator."""
        return DiagonalOperator(self._values.conjugate())

    def to_dense(self):
        """Return dense matrix."""
        return _np.diag(self._values)

    def to_sparse(self):
        """Return sparse matrix if operator is sparse."""
        from scipy.sparse import diags

        return diags(self._values)


class SparseDiscreteBoundaryOperator(_DiscreteOperatorBase):
    """
    Main class for the discrete form of sparse operators.

    This class derives from
    :class:`scipy.sparse.linalg.interface.LinearOperator`
    and thereby implements the SciPy LinearOperator protocol.

    """

    def __init__(self, impl):
        """Construct. Should not be called by the user."""
        super(SparseDiscreteBoundaryOperator, self).__init__(impl.dtype, impl.shape)
        self._impl = impl

    def _matmat(self, vec):
        """Multiply the operator with a numpy vector or matrix x."""
        if self.dtype == "float64" and _np.iscomplexobj(vec):
            return self.to_sparse() * _np.real(vec) + 1j * (self.to_sparse() * _np.imag(vec))
        return self.to_sparse() * vec

    def _transpose(self):
        """Return the transpose of the discrete operator."""
        return SparseDiscreteBoundaryOperator(self.to_sparse().transpose())

    def _adjoint(self):
        """Return the adjoint of the discrete operator."""
        return SparseDiscreteBoundaryOperator(self.to_sparse().transpose().conjugate())

    def __add__(self, other):
        """Add."""
        if isinstance(other, SparseDiscreteBoundaryOperator):
            return SparseDiscreteBoundaryOperator(self.to_sparse() + other.to_sparse())
        else:
            return super().__add__(other)

    def __neg__(self):
        """Negate."""
        return SparseDiscreteBoundaryOperator(-self.to_sparse())

    def __mul__(self, other):
        """Multiply."""
        if isinstance(other, SparseDiscreteBoundaryOperator):
            return SparseDiscreteBoundaryOperator(self.to_sparse() * other.to_sparse())
        else:
            return super().__mul__(other)

    def dot(self, other):
        """Product with other objects."""
        if _np.isscalar(other):
            return SparseDiscreteBoundaryOperator(self.to_sparse() * other)
        else:
            return super().dot(other)

    def __rmul__(self, other):
        """Multiply."""
        if _np.isscalar(other):
            return SparseDiscreteBoundaryOperator(self.to_sparse() * other)
        else:
            return NotImplemented

    @property
    def A(self):
        """Return dense matrix."""
        warnings.warn(
            "operator.A is deprecated and will be removed in a future version. Use operator.to_dense() instead.",
            DeprecationWarning,
        )
        return self.to_sparse()

    def to_dense(self):
        """Return dense matrix."""
        return self._impl.todense()

    def to_sparse(self):
        """Return sparse matrix if operator is sparse."""
        return self._impl


class InverseSparseDiscreteBoundaryOperator(_DiscreteOperatorBase):
    """
    Apply the (pseudo-)inverse of a sparse operator.

    This class uses a Sparse LU-Decomposition
    (in the case of a square matrix) or a sparse normal
    equation to provide the application of an inverse to
    a sparse operator.

    This class derives from
    :class:`scipy.sparse.linalg.interface.LinearOperator`
    and thereby implements the SciPy LinearOperator protocol.

    Parameters
    ----------
    operator : bempp_cl.api.SparseDiscreteBoundaryOperator
        Sparse operator to be inverted.

    """

    def __init__(self, operator):
        """Create a inverse sparse operator."""
        self._solver = _Solver(operator)
        self._operator = operator
        super().__init__(self._solver.dtype, self._solver.shape)

    def _matmat(self, vec):
        """Implemententation of matvec."""
        return self._solver.solve(vec)

    def to_dense(self):
        """Return dense matrix."""
        eye = _np.eye(self.shape[1])
        return self @ eye

    def to_sparse(self):
        """Return sparse matrix if operator is sparse."""
        raise ValueError("This matrix is not sparse.")


class ZeroDiscreteBoundaryOperator(_DiscreteOperatorBase):
    """A discrete operator that represents a zero operator.

    This class derives from
    :class:`scipy.sparse.linalg.interface.LinearOperator`
    and thereby implements the SciPy LinearOperator protocol.

    Parameters
    ----------
    rows : int
        The number of rows in the operator.
    columns : int
        The number of columns in the operator.

    """

    def __init__(self, rows, columns):
        """Construct a zero operator."""
        super(ZeroDiscreteBoundaryOperator, self).__init__(_np.dtype("float64"), (rows, columns))

    def _matmat(self, x):
        """Multiply operator by a matrix."""
        return _np.zeros((self.shape[0], x.shape[1]), dtype="float64")

    def to_dense(self):
        """Return dense matrix."""
        return _np.zeros(self.shape)

    def to_sparse(self):
        """Return sparse matrix if operator is sparse."""
        from scipy.sparse import csc_matrix

        return csc_matrix(self.shape, dtype="float64")


class DiscreteRankOneOperator(_DiscreteOperatorBase):
    """Creates a discrete rank one operator.

    This class represents a rank one operator given
    by column * row, where column is column is
    interpreted as a (m, 1) array and row as
    a (1, n) array.

    Parameters
    ----------
    column : np.array
        A column vector
    row : np.array
        A row vector

    """

    def __init__(self, column, row):
        """Construct a discrete rank one operator."""
        if row.dtype == "complex128" or column.dtype == "complex128":
            dtype = "complex128"
        else:
            dtype = "float64"

        self._row = row.ravel()
        self._column = column.ravel()

        shape = (len(self._column), len(self._row))
        super().__init__(dtype, shape)

    def _matvec(self, x):
        """Multiply operator by a vector."""
        if x.ndim > 1:
            return _np.outer(self._column, _np.dot(self._row, x))
        else:
            return self._column * _np.dot(self._row, x)

    def _rmatvec(self, x):
        """Multiply."""
        # pylint: disable=protected-access
        return self._adjoint()._matvec(x)

    def _transpose(self):
        """Find the transpose."""
        return DiscreteRankOneOperator(self._row, self._column)

    def _adjoint(self):
        """Find the adjoint."""
        return DiscreteRankOneOperator(self._row.conjugate(), self._column.conjugate())

    def to_dense(self):
        """Return dense matrix."""
        return _np.outer(self._column, self._row)

    def to_sparse(self):
        """Return sparse matrix if operator is sparse."""
        raise ValueError("This matrix is not sparse.")


def as_matrix(operator):
    """
    Convert a discrte operator into a dense matrix.

    Parameters
    ----------
    operator : scipy.sparse.linalg.interface.LinearOperator
        The linear operator to be converted into a matrix.


    Notes
    -----
    Note that this function may be slow depending on how the original
    discrete operator was stored. In the case of a dense assembly simple
    the underlying NumPy matrix is returned. For sparse matrices the
    corresponding Scipy sparse matrix is returned. Otherwise, the operator needs
    to be converted to an array, which can take a long time depending on
    the assembler type.

    """
    return operator.to_dense()


class _Solver(object):  # pylint: disable=too-few-public-methods
    """Actual solve of a sparse linear system."""

    # pylint: disable=too-many-locals
    @_timeit
    def __init__(self, operator):
        """Create a solver."""
        from scipy.sparse import csc_matrix

        if isinstance(operator, SparseDiscreteBoundaryOperator):
            mat = operator.to_sparse()
        elif isinstance(operator, csc_matrix):
            mat = operator
        else:
            raise ValueError(
                "op must be either of type "
                + "SparseDiscreteBoundaryOperator or of type "
                + "csc_matrix. Actual type: "
                + str(type(operator))
            )

        from scipy.sparse.linalg import splu

        self._solve_fun = None
        self._shape = (mat.shape[1], mat.shape[0])
        self._dtype = mat.dtype

        use_mkl_pardiso = False

        # pylint: disable=bare-except
        try:
            # pylint: disable=E0401
            from mkl_pardiso_solve import PardisoInterface

            # pylint: disable=invalid-name
            solver_interface = PardisoInterface
            actual_mat = mat.tocsr()
            use_mkl_pardiso = True
        except:  # noqa: E722
            solver_interface = splu
            actual_mat = mat

        if mat.shape[0] == mat.shape[1]:
            # Square matrix case
            solver = solver_interface(actual_mat)
            self._solve_fun = solver.solve
        elif mat.shape[0] > mat.shape[1]:
            # Thin matrix case
            mat_hermitian = actual_mat.conjugate().transpose()
            if use_mkl_pardiso:
                solver = solver_interface((mat_hermitian * mat).tocsr())
            else:
                solver = solver_interface((mat_hermitian * mat).tocsc())
            self._solve_fun = lambda x: solver.solve(mat_hermitian * x)
        else:
            # Thick matrix case

            mat_hermitian = actual_mat.conjugate().transpose()
            if use_mkl_pardiso:
                solver = solver_interface((mat * mat_hermitian).tocsr())
            else:
                solver = solver_interface((mat * mat_hermitian).tocsc())
            self._solve_fun = lambda x: mat_hermitian * solver.solve(x)

    @_timeit
    def solve(self, rhs):
        """Solve with right-hand side mat."""
        if self._dtype == "float64" and _np.iscomplexobj(rhs):
            return self.solve(_np.real(rhs)) + 1j * self.solve(_np.imag(rhs))

        return self._solve_fun(rhs)

    @property
    def shape(self):
        """Return the shape of the inverse operator."""
        return self._shape

    @property
    def dtype(self):
        """Return the dtype."""
        return self._dtype



"""Define various test shapes."""

from .shapes import regular_sphere
from .shapes import cube
from .shapes import multitrace_cube
from .shapes import reference_triangle
from .shapes import sphere
from .shapes import ellipsoid
from .shapes import cuboid
from .shapes import almond
from .shapes import reentrant_cube
from .shapes import screen
from .shapes import cylinders
from .shapes import multitrace_ellipsoid
from .shapes import multitrace_sphere



"""Implementation of MPI based remote operators."""

from bempp_cl.api.assembly.blocked_operator import BlockedOperator as _BlockedOperator
from bempp_cl.api.assembly.discrete_boundary_operator import _DiscreteOperatorBase


from mpi4py import MPI
import numpy as _np

MPI_SIZE = MPI.COMM_WORLD.Get_size()
MPI_RANK = MPI.COMM_WORLD.Get_rank()

COMM = MPI.COMM_WORLD

_REMOTE_MANAGER = None


class Message:
    """Messages for remote workers."""

    def __init__(self, status, operator_tag=None, is_complex=False, nelements=None):
        """Create message."""
        self.status = status
        self.is_complex = is_complex
        self.operator_tag = operator_tag
        self.nelements = nelements


class RemoteManager:
    """Manage remote worker execution."""

    def __init__(self):
        """Create manager."""
        self._op_data = {}
        self._op_location = {}
        self._tags = {}
        self._tags_counter = 0
        self._rank_counter = 1

    def register(self, op):
        """Register operator for remote execution."""
        tag = self._tags_counter
        rank = self._rank_counter

        self._op_data[tag] = op
        self._op_location[tag] = rank
        self._tags[op] = tag

        self._tags_counter += 1
        self._rank_counter = (1 + self._rank_counter) % MPI_SIZE
        if self._rank_counter == 0:
            self._rank_counter = 1

    def execute_worker(self):
        """Execute on workers."""
        while True:
            msg, tag, data = self.receive_data(0)

            if msg == "SHUTDOWN":
                break
            elif msg == "ASSEMBLE":
                self._op_data[tag].weak_form()
            elif msg == "SYNCHRONIZE":
                self.send_data("SYNCHRONIZED", 0, tag)
            elif msg == "GET_DTYPE":
                self.send_data(_np.dtype(self._op_data[tag].weak_form().dtype).name, 0, tag)
            elif msg == "DATA":
                result = self._op_data[tag].weak_form() @ data
                self.send_data("DATA", 0, tag, result)

    def send_data(self, msg, dest, operator_tag=None, data=None):
        """Send data to destination rank."""
        COMM.send(
            Message(
                msg,
                operator_tag=operator_tag,
                is_complex=_np.iscomplexobj(data),
                nelements=len(data) if data is not None else None,
            ),
            dest=dest,
        )
        if msg == "DATA":
            COMM.Send(data, dest=dest)

    def send_error(self, msg, operator_tag):
        """Send error message to master."""
        COMM.send(Message(msg, operator_tag=operator_tag), dest=0)

    def receive_data(self, source):
        """Receive data."""
        message = COMM.recv(source=source)

        if message.status == "SHUTDOWN":
            return message.status, None, None
        elif message.status == "DATA":
            if message.is_complex:
                dtype = _np.complex128
            else:
                dtype = _np.float64
            data = _np.empty(message.nelements, dtype=dtype)
            COMM.Recv(data, source=source)
            return message.status, message.operator_tag, data
        else:
            return message.status, message.operator_tag, None

    def submit_computation(self, op, x):
        """Submit computation to operator."""
        tag = self._tags[op]
        rank = self._op_location[tag]

        self.send_data("DATA", rank, tag, x)

    def receive_result(self, op):
        """Receive result from a specific operator."""
        tag = self._tags[op]
        rank = self._op_location[tag]
        _, _, data = self.receive_data(rank)
        return data

    def assemble(self, op):
        """Assemble a given remote operator."""
        tag = self._tags[op]
        rank = self._op_location[tag]
        self.send_data("ASSEMBLE", rank, tag)

    def barrier(self):
        """Use barrier operation for workers."""
        for rank in range(1, MPI_SIZE):
            self.send_data("SYNCHRONIZE", rank, 0)
            msg, _, _ = self.receive_data(rank)
            if msg != "SYNCHRONIZED":
                raise Exception(f"Error: expected message 'SYNCHRONIZED' from rank {0}, received {msg}")

    def get_operator_dtype(self, tag):
        """Get dtype of remote operator."""
        rank = self._op_location[tag]
        self.send_data("GET_DTYPE", rank, tag)
        msg, _, _ = self.receive_data(rank)
        return _np.dtype(msg)

    def shutdown(self):
        """Shutdown all workers."""
        if MPI_SIZE == 0:
            return

        for worker in range(1, MPI_SIZE):
            self.send_data("SHUTDOWN", worker)

    def _group_by_rank(self, tags):
        """
        Group tags by for parallel computation.

        Given a list of tags (with possible repetitions).
        Return a list of list of positions in the tags list
        that is sorted so that each sublist does not have repeating
        ranks. This allows to distribute the computations effectively
        on the ranks.
        """
        output = []

        tags_with_indices = {index: tag for index, tag in enumerate(tags)}

        while len(tags_with_indices) > 0:
            current_indices = []
            current_ranks = []
            for index, tag in tags_with_indices.items():
                rank = self._op_location[tag]
                if rank not in current_ranks:
                    current_indices.append(index)
                    current_ranks.append(rank)
            for index in current_indices:
                del tags_with_indices[index]
            output.append(current_indices)

        return output

    def assemble_parallel(self, ops):
        """Assemble a list of operators in parallel."""
        tags = [self._tags[op] for op in ops]
        grouped_indices = self._group_by_rank(tags)

        for index_list in grouped_indices:
            for index in index_list:
                self.assemble(ops[index])
            self.barrier()

    def compute_parallel(self, tasks):
        """
        Compute in parallel.

        Parameters
        ----------
        tasks : list
            A list [(op1, data1), (op2, data2), ...]
            of operators and associated vectors.

        Returns a list [result1, result2, ...]
        of the results of the computation.

        Operators are allowed to appear repeatedly in the list
        and are scheduled correctly.

        """
        tags = [self._tags[op] for op, _ in tasks]

        results = len(tasks) * [None]
        grouped_indices = self._group_by_rank(tags)

        for index_list in grouped_indices:
            for index in index_list:
                op, data = tasks[index]
                self.submit_computation(op, data)
            for index in index_list:
                op, data = tasks[index]
                results[index] = self.receive_result(op)

        return results

    @property
    def tags(self):
        """Return tags."""
        return self._tags


def get_remote_manager():
    """Initialize remote manager."""
    global _REMOTE_MANAGER

    if _REMOTE_MANAGER is None:
        _REMOTE_MANAGER = RemoteManager()
    return _REMOTE_MANAGER


class RemoteBlockedOperator(_BlockedOperator):
    """Define a remote blocked operator."""

    def __init__(self, m, n):
        """Initialize an m x n remote blocked operator."""
        super().__init__(m, n)

    def _assemble(self):
        """Assemble the operator."""
        if not self._fill_complete():
            raise ValueError("Each row and column must have at least one operator")

        return RemoteBlockedDiscreteOperator(self._operators)


class RemoteBlockedDiscreteOperator(_DiscreteOperatorBase):
    """Implementation of a discrete blocked boundary operator."""

    def __init__(self, ops):
        """
        Construct an operator from a two dimensional Numpy array of operators.

        ops is a list of list containing discrete boundary operators or None.
        A None entry is equivalent to a zero discrete boundary operator.

        """
        # pylint: disable=too-many-branches
        self._manager = get_remote_manager()

        if not isinstance(ops, _np.ndarray):
            ops = _np.array(ops)

        rows = ops.shape[0]
        cols = ops.shape[1]

        self._ndims = (rows, cols)

        self._operators = _np.empty((rows, cols), dtype=object)
        self._tags = _np.empty((rows, cols), dtype=_np.int)
        self._rows = -_np.ones(rows, dtype=int)
        self._cols = -_np.ones(cols, dtype=int)

        for i in range(rows):
            for j in range(cols):
                if ops[i, j] is None:
                    continue
                if self._rows[i] != -1:
                    if self._get_shape(ops[i, j])[0] != self._rows[i]:
                        raise ValueError("Block row {0} has incompatible ".format(i) + " operator sizes.")
                else:
                    self._rows[i] = self._get_shape(ops[i, j])[0]

                if self._cols[j] != -1:
                    if self._get_shape(ops[i, j])[1] != self._cols[j]:
                        raise ValueError("Block column {0} has incompatible".format(j) + "operator sizes.")
                else:
                    self._cols[j] = self._get_shape(ops[i, j])[1]
                self._operators[i, j] = ops[i, j]

        if not self._fill_complete():
            raise ValueError("Each row and column must contain at least one operator.")

        for i in range(rows):
            for j in range(cols):
                if self._operators[i, j] is None:
                    self._tags[i, j] = -1
                else:
                    self._tags[i, j] = self._manager.tags[self._operators[i, j]]

        shape = (_np.sum(self._rows), _np.sum(self._cols))

        is_complex = False
        for i in range(rows):
            for j in range(cols):
                if self._tags[i, j] == -1:
                    continue
                self._manager.assemble(self._operators[i, j])

        self._manager.barrier()

        for i in range(rows):
            for j in range(cols):
                if self._tags[i, j] == -1:
                    continue
                dtype = self._manager.get_operator_dtype(self._tags[i, j])
                if dtype in ["complex128", "complex64"]:
                    is_complex = True

        if is_complex:
            dtype = _np.complex128
        else:
            dtype = _np.float64

        super().__init__(dtype, shape)

    def __getitem__(self, key):
        """Return the object at position (i, j)."""
        return self._operators[key]

    def _get_shape(self, op):
        """Get shape of boundary operator."""
        return (op.dual_to_range.global_dof_count, op.domain.global_dof_count)

    def _fill_complete(self):
        """Check if the fill is complete."""
        if (-1 in self._rows) or (-1 in self._cols):
            return False
        return True

    def _matvec(self, x):
        """Perform matrix vector multiplication."""
        from bempp_cl.api.utils.data_types import combined_type

        if not self._fill_complete():
            raise ValueError("Not all rows or columns contain operators.")

        ndims = len(x.shape)
        x = x.ravel()

        row_dim = 0
        res = _np.zeros(self.shape[0], dtype=combined_type(self.dtype, x.dtype))

        compute_tasks = []

        # Submit the computations
        for i in range(self._ndims[0]):
            col_dim = 0
            for j in range(self._ndims[1]):
                if self._tags[i, j] != -1:
                    # If self._tags[i, j] == -1 the operator is None
                    local_x = x[col_dim : col_dim + self._cols[j]]
                    compute_tasks.append((self._operators[i, j], local_x))
                col_dim += self._cols[j]
            row_dim += self._rows[i]

        # Execute computations

        compute_results = self._manager.compute_parallel(compute_tasks)

        # Get results back
        row_dim = 0
        task_count = 0
        for i in range(self._ndims[0]):
            col_dim = 0
            for j in range(self._ndims[1]):
                if self._tags[i, j] == -1:
                    continue
                res[row_dim : row_dim + self._rows[i]] += compute_results[task_count]
                task_count += 1
            row_dim += self._rows[i]

        if ndims == 2:
            res = res.reshape(-1, 1)

        return res

    def _get_row_dimensions(self):
        """Return the number of rows."""
        return self._rows

    def _get_column_dimensions(self):
        """Return the number of columns."""
        return self._cols

    @property
    def A(self):
        """Return as dense Numpy array."""
        raise NotImplementedError()

    def _transpose(self):
        """Implement the transpose."""
        raise NotImplementedError()

    def _adjoint(self):
        """Implement the adjoint."""
        raise NotImplementedError()

    row_dimensions = property(_get_row_dimensions)
    column_dimensions = property(_get_column_dimensions)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Oct  3 15:47:02 2023

@author: tom.munoz
"""
import numpy as np
from os import listdir
from os.path import isfile, join
import re

def load_impulse(folderPath, isConvertedFromMAT=False):
    """
    Import impulse response data from CSV files.

    Parameters
    ----------
    folderPath : str
        Path to the folder where IR.csv files are stored.

    isConvertedFromMAT : bool, optional
        If True, indicates that the data has been converted from .mat to .csv format,
        which may affect the number of rows to skip when loading data. Default is False.

    Returns
    -------
    theta : numpy array
        Array of angles in degrees corresponding to the imported impulse responses.

    IR_MATRIX : numpy array
        Impulse response matrix of shape (len(theta), len(IR)).

    Notes
    -----
    This function imports impulse response data from CSV files stored in the specified
    'folderPath'. The function returns a tuple containing two numpy arrays: 'theta' and
    'IR_MATRIX'. The 'theta' array holds the angles in degrees corresponding to the
    imported impulse responses. The 'IR_MATRIX' array holds the actual impulse response
    data in a matrix format, where each row represents an angle and each column represents
    a time step in the impulse response.

    If the 'isConvertedFromMAT' parameter is set to True, the function adjusts the number of
    rows to skip when loading data to accommodate the conversion from .mat to .csv format.
    """
    files = [f for f in listdir(folderPath) if isfile(join(folderPath, f))]
    files = [f for f in files if "IR" in f]

    # load file to get length of time data
    if isConvertedFromMAT is False:
        dummyData = np.loadtxt(join(folderPath, files[0]), skiprows=4, delimiter=',')
        lenData = len(dummyData[:, 0])
        skiprow = 4
    elif isConvertedFromMAT is True:
        dummyData = np.loadtxt(join(folderPath, files[0]), skiprows=1, delimiter=',')
        lenData = len(dummyData[:, 0])
        skiprow = 1

    # build angle array
    pattern = r'[\d\.]+'
    numerical_values = re.findall(pattern, files[0][:-4])

    where_theta = files[0].index(numerical_values[0])

    theta = np.zeros(len(files))
    for t in range(len(theta)):
        theta[t] = float(files[t][where_theta:where_theta + 3])
    theta = np.sort(theta)

    # build IR matrix
    IR_MATRIX = np.zeros([len(theta), lenData])
    for t in range(len(theta)):
        filepath = join(folderPath, files[0][:where_theta] + str(theta[t]) + '_' + 'IR.csv')
        data = np.loadtxt(filepath, skiprows=skiprow, delimiter=',')
        time = data[:, 0]
        signal = data[:, 1]
        IR_MATRIX[t, :] = signal
    return theta, IR_MATRIX



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Sep  3 10:40:49 2024

@author: tom
"""

from numpy import array, zeros, ones

class CCVS:
    def __init__(self, np, nm, np1, nm1, value):
        """
        Create a gyrator component

        Parameters
        ----------
        np : TYPE
            DESCRIPTION.
        nm : TYPE
            DESCRIPTION.
        np1 : TYPE
            DESCRIPTION.
        nm1 : TYPE
            DESCRIPTION.
        value : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        
        self.np = str(np)
        self.nm = str(nm)
        self.np1 = str(np1)
        self.nm1 = str(nm1)
        self.value = value
        self.G = 1
        self.Gs = None
        
        # create stamp and relative informations
        self.stamp_G = array([[0, 0, 1], 
                              [0, 0, -1], 
                              [1, -1, 0]])
        self.stamp_I = None
        self.contribute = ["G", "Is"]  
        self.vsource = 2
        
    
    def init_component(self, frequency):
        self.Gs = ones(len(frequency))
    
    def update_stamp(self, node_id, M, nbsource):
        """
        Update the component's stamp in a matrix of the system's size.

        Parameters
        ----------
        node_list : list
            list of nodes.

        Returns
        -------
        None.

        """
        if self.np != '0':
            np = node_id[self.np]
        else:
            np = 0
            
        if self.nm != '0':
            nm = node_id[self.nm]
        else:
            nm = 0
        
        if self.np1 != '0':
            np1 = node_id[self.np1]
        else:
            np1 = 0
            
        if self.nm1 != '0':
            nm1 = node_id[self.nm1]
        else:
            nm1 = 0
        
        maxNode = max(node_id.values())

        # G stamp
        self.stamp_G = zeros([maxNode+M, maxNode+M], dtype=complex)
        
        # current sensor
        if np != 0:
            self.stamp_G[np-1, maxNode+nbsource] = 1
            self.stamp_G[maxNode+nbsource, np-1] = 1
        if nm != 0:
            self.stamp_G[nm-1, maxNode+nbsource] = -1
            self.stamp_G[maxNode+nbsource, nm-1] = -1        
        
        # secondary source
        if np1 != 0:
            self.stamp_G[np1-1, maxNode+nbsource+1] = 1
            self.stamp_G[maxNode+nbsource+1, np1-1] = 1
        if nm1 != 0:
            self.stamp_G[nm1-1, maxNode+nbsource+1] = -1
            self.stamp_G[maxNode+nbsource+1, nm1-1] = -1 
        
        # Bl factor
        self.stamp_G[maxNode+nbsource+1, maxNode+nbsource] = -self.value
                
        # Y stamp
        self.stamp_I = zeros([maxNode+M, 1], dtype=complex)
        
        